#
# Functions relating to the DataLog class for storing time series data.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os
import re
import sys
import array

import numpy as np

from collections import OrderedDict

import myokit


# Function to split keys into dimension-key,qname-key pairs
ID_NAME_PATTERN = re.compile(r'(\d+.)+')

# Readme file for DataLog binary files
README_SAVE_BIN = """
Myokit DataLog Binary File
--------------------------
This zip file contains binary time series data for one or multiple variables.
The file structure.txt contains structural information about the data in plain
text. The first line lists the number of fields. The second line gives the
length of the data arrays. The third line specifies the data type, either
single ("f") or double ("d") precision. The fourth line indicates which entry
corresponds to a time variable, or is blank if no time variable was explicitly
specified. Each following line contains the name of a data field, in the order
its data occurs in the binary data file "data.bin". All data is stored
little-endian.
""".strip()

# Encoding used for text portions of zip files
ENC = 'utf-8'


class DataLog(OrderedDict):
    """
    A dictionary time-series, for example data logged during a simulation or
    experiment.

    A :class:`DataLog` is expected but not required to contain a single
    entry indicating time and any number of entries representing a variable
    varying over time.

    Single cell data is accessed simply by the variable name::

        v = log['membrane.V']

    Multi-cell data is accessed by appending the index of the cell before the
    variable name. For example::

        v = log['1.2.membrane.V']

    This returns the membrane potential for cell (1, 2). Another way to obtain
    the same result is::

        v = log['membrane.V', 1, 2]

    or, finally::

        v = log['membrane.V', (1, 2)]

    Every array stored in the log must have the same length. This condition can
    be checked by calling the method :meth:`validate`.

    A new ``DataLog`` can be created in a number of ways::

        # Create an empty DataLog:
        d = myokit.DataLog()
        d['time'] = [1, 2, 3]
        d['data'] = [2, 4, 5]
        d.set_time_key('time')

        # Create a clone of d
        e = myokit.DataLog(d)

        # Create a DataLog based on a dictionary
        d = myokit.DataLog({'time':[1, 2, 3], 'data':[2, 4, 5]}, time='time')

    Arguments:

    ``other``
        A DataLog to clone or a dictionary to use as basis.
    ``time``
        The log key to use for the time variable. When cloning a log, adding
        the ``time`` argument will overwrite the cloned value.

    """

    def __init__(self, other=None, time=None):
        if other is None:
            # Create new
            super().__init__()
            self._time = None
        else:
            # Clone
            super().__init__(other)
            try:
                self._time = str(other._time)
            except Exception:
                self._time = None
        if time is not None:
            self.set_time_key(time)

    def apd(self, v='membrane.V', threshold=-70):
        """
        Calculates one or more Action Potential Durations (APDs) in a single
        cell's membrane potential.

        *Note 1: More accuracte APD measurements can be obtained using the*
        :class:`Simulation` *object's APD tracking functionality. See*
        :meth:`Simulation.run()` *for details.*

        *Note 2: This APD is defined by simply checking crossing of a threshold
        potential, and does not look at lowest or highest voltages in a
        signal.*

        The membrane potential data should be listed in the log under the key
        given by ``v``.

        The returned value is a list of tuples (AP_start, APD).
        """
        def crossings(x, y, t):
            """
            Calculates the ``x``-values where ``y`` crosses threshold ``t``.
            Returns a list of tuples ``(xc, sc)`` where ``xc`` is the ``x``
            coordinate of the crossing and ``sc`` is the slope at this point.
            """
            x = np.asarray(x)
            y = np.asarray(y)

            # Get array of places where v exceeds the threshold (1s and 0s)
            h = (y > t) * 1

            # Get array of indices just before a crossing
            c = np.argwhere(h[1:] - h[:-1])

            # Gather crossing times
            crossings = []
            for i in c:
                i = i[0]
                sc = (y[i + 1] - y[i]) / (x[i + 1] - x[i])
                if y[i] == t:
                    xc = x[i]
                else:
                    xc = x[i] + (t - y[i]) / sc
                crossings.append((xc, sc))
            return crossings

        # Check time variable
        t = np.asarray(self.time())

        # Check voltage variable
        v = np.asarray(self[v])

        # Check threshold
        threshold = float(threshold)

        # Initial status: check if already in AP
        apds = myokit.DataLog()
        apds['start'] = []
        apds['duration'] = []
        last = t[0]

        # Evaluate crossings
        for time, slope in crossings(t, v, threshold):
            if slope < 0:
                # End of AP
                if last != t[0]:    # Don't inlcude AP started before t[0]
                    apds['start'].append(last)
                    apds['duration'].append(time - last)
            last = time
        return apds

    def block1d(self):
        """
        Returns a copy of this log as a :class:`DataBlock1d`.
        """
        return myokit.DataBlock1d.from_log(self)

    def block2d(self):
        """
        Returns a copy of this log as a :class:`DataBlock2d`.
        """
        return myokit.DataBlock2d.from_log(self)

    def clone(self, numpy=False):
        """
        Returns a deep clone of this log.

        All lists in the log will be duplicated, but the list contents are
        assumed to be numerical (and thereby immutable) and won't be cloned.

        A log with numpy arrays instead of lists can be created by setting
        ``numpy=True``.
        """
        log = DataLog()
        log._time = self._time
        if numpy:
            for k, v in self.items():
                log[str(k)] = np.array(v, copy=True, dtype=float)
        else:
            for k, v in self.items():
                log[str(k)] = list(v)
        return log

    def __contains__(self, key):
        return super().__contains__(self._parse_key(key))

    def __delitem__(self, key):
        return super().__delitem__(self._parse_key(key))

    def extend(self, other):
        """
        Returns a copy of this log, extended with the data of another.

        Both logs must have the same keys and the same time key. The added data
        must be from later time points than in the log being extended.
        """
        if other._time != self._time:
            raise ValueError('Both logs must have the same time key.')
        if other[self._time][0] < self[self._time][-1]:
            raise ValueError(
                'Cannot extend DataLog with data from an earlier time.')
        if set(self.keys()) != set(other.keys()):
            raise ValueError('Both logs must have the same keys.')

        # Create new log
        log = DataLog()
        log._time = self._time

        # Add data
        for k, v1 in self.items():
            v2 = other[k]
            if isinstance(v1, np.ndarray) or isinstance(v2, np.ndarray):
                # Concatenation copies data
                log[k] = np.concatenate((np.asarray(v1), np.asarray(v2)))
            else:
                log[k] = list(v1)   # Copies v1 data
                log[k].extend(v2)   # Copies v2 data

        # Return
        return log

    def find(self, time):
        """
        Deprecated alias of :meth:`find_after()`.
        """
        # Deprecated since 2019-08-16
        import warnings
        warnings.warn(
            'The method `find` is deprecated. Please use `find_after`'
            ' instead.')

        return self.find_after(time)

    def find_after(self, time):
        """
        Returns the lowest index ``i`` such that

            times[i] >= time

        where ``times`` are the times stored in this ``DataLog``.

        If no such value exists in the log, ``len(time)`` is returned.
        """
        times = self[self._time]

        # Border cases
        n = len(times)
        if n == 0 or time <= times[0]:
            return 0
        if time > times[-1]:
            return n

        # Find t
        def find(lo, hi):
            # lo = first index, hi = last index + 1
            if (lo + 1 == hi):
                return lo + 1
            m = int((lo + hi) / 2)
            if time > times[m]:
                return find(m, hi)
            else:
                return find(lo, m)

        return find(0, n)

    def fold(self, period, discard_remainder=True):
        """
        Creates a copy of the log, split with the given period. Split signals
        are given indexes so that "current" becomes "0.current", "1.current"
        "2.current", etc.

        If the logs entries do not divide well by 'period', the remainder will
        be ignored. This happens commonly due to rounding point errors (in
        which case the remainder is a single entry). To disable this behavior,
        set ``discard_remainder=False``.
        """
        # Note: Using closed intervals can lead to logs of unequal length, so
        # it should be disabled here to ensure a valid log
        logs = self.split_periodic(period, adjust=True, closed_intervals=False)
        # Discard remainder if present
        if discard_remainder:
            if len(logs) > 1:
                n = logs[0].length()
                if logs[-1].length() < n:
                    logs = logs[:-1]
        # Create new log with folded data
        out = myokit.DataLog()
        out._time = self._time
        out[self._time] = logs[0][self._time]
        for i, log in enumerate(logs):
            for k, v in log.items():
                if k != self._time:
                    out[k, i] = v
        return out

    def __getitem__(self, key):
        return super().__getitem__(self._parse_key(key))

    def has_nan(self):
        """
        Returns True if one of the variables in this DataLog has a ``NaN`` as
        its final logged value.
        """
        for k, d in self.items():
            if len(d) > 0 and np.isnan(d[-1]):
                return True
        return False

    def integrate(self, name, *cell):
        """
        Integrates a field from this log and returns it::

            # Run a simulation and calculate the total current carried by INa
            s = myokit.Simulation(m, p)
            d = s.run(1000)
            q = d.integrate('ina.INa')

        Arguments:

        ``name``
            The name of the variable to return, for example 'ik1.IK1' or
            '2.1.membrane.V'.
        ``*cell``
            An optional cell index, for easy access to multi-cellular data, for
            example ``log.integrate('membrane.V', 2, 1)``.

        """
        # Get data to integrate
        key = [str(x) for x in cell]
        key.append(str(name))
        key = '.'.join(key)
        data = np.array(self[key], copy=True)
        time = np.asarray(self.time())

        # Integration using the midpoint Riemann sum:
        #  At point i=0, the value is 0
        #  At each point i>0, the value increases by step[i, i-1]*mean[i, i-1]
        #data[1:] = 0.5 * (data[1:] + data[:-1]) * (time[1:] - time[:-1])
        #data[0]  = 0

        # For discontinuities (esp. with CVODES), it makes more sense to treat
        # the signal as a zero-order hold, IE use the left-point integration
        # rule:
        data[1:] = data[:-1] * (time[1:] - time[:-1])
        data[0] = 0
        return data.cumsum()

    def interpolate_at(self, name, time):
        """
        Returns the value for variable ``name`` at a given ``time``, determined
        using linear interpolation between the nearest matching times.
        """
        t = self[self._time]
        v = self[name]

        # Don't extrapolate
        if time < t[0] or time > t[-1]:
            raise ValueError(
                'Requested time is outside of logged range, would require'
                ' extrapolation.')

        # Get first time *after or at* requested time
        i1 = self.find_after(time)

        # Return directly, if possible
        if t[i1] == time:
            return v[i1]

        # Interpolate
        i0 = i1 - 1
        return v[i0] + (time - t[i0]) * (v[i1] - v[i0]) / (t[i1] - t[i0])

    def isplit(self, i):
        """
        Returns two logs, where the first contains all this log's entries up to
        index ``i``, and the second contains all entries starting from ``i``
        and higher.
        """
        log1 = DataLog()
        log2 = DataLog()
        log1._time = self._time
        log2._time = self._time
        for k, v in self.items():
            if isinstance(v, np.ndarray):
                log1[k] = np.array(v[:i], copy=True, dtype=float)
                log2[k] = np.array(v[i:], copy=True, dtype=float)
            else:
                log1[k] = v[:i]
                log2[k] = v[i:]
        return log1, log2

    def itrim(self, a, b):
        """
        Returns a copy of this log, with all entries trimmed to the region
        between indices ``a`` and ``b`` (similar to performing ``x = x[a:b]``
        on a list).
        """
        log = DataLog()
        log._time = self._time
        for k, v in self.items():
            if isinstance(v, np.ndarray):
                log[k] = np.array(v[a:b], copy=True, dtype=float)
            else:
                log[k] = v[a:b]
        return log

    def itrim_left(self, i):
        """
        Returns a copy of this log, with all entries before index ``i``
        removed (similar to performing ``x = x[i:]`` on a list).
        """
        log = DataLog()
        log._time = self._time
        for k, v in self.items():
            if isinstance(v, np.ndarray):
                log[k] = np.array(v[i:], copy=True, dtype=float)
            else:
                log[k] = v[i:]
        return log

    def itrim_right(self, i):
        """
        Returns a copy of this log, with all entries starting from index ``i``
        removed (similar to performing ``x = x[:i]`` on a list).
        """
        log = DataLog()
        log._time = self._time
        for k, v in self.items():
            if isinstance(v, np.ndarray):
                log[k] = np.array(v[:i], copy=True, dtype=float)
            else:
                log[k] = v[:i]
        return log

    def keys_like(self, query):
        """
        Returns all keys that match the pattern ``*.query``, sorted
        alphabetically.

        For example, ``log.keys_like('membrane.V')`` could return a list
        ``['0.membrane.V', '1.membrane.V', '2.membrane,V', ...]``, or
        ``['0.0.membrane.V', '0.1.membrane.V', '1.0.membrane,V', ...]``.
        """
        keys = [x for x in self.keys() if x.endswith('.' + str(query))]
        keys.sort()
        return keys

    def length(self):
        """
        Returns the length of the entries in this log. If the log is empty,
        zero is returned.
        """
        if len(self) == 0:
            return 0
        return len(next(iter(self.values())))

    @staticmethod
    def load(filename, progress=None, msg='Loading DataLog'):
        """
        Loads a :class:`DataLog` from the binary format used by myokit.

        The values in the log will be stored in an :class:`array.array`. The
        data type used by the array will be the one specified in the binary
        file. Notice that an `array.array` storing single precision floats will
        make conversions to ``Float`` objects when items are accessed.

        To obtain feedback on the simulation progress, an object implementing
        the :class:`myokit.ProgressReporter` interface can be passed in.
        passed in as ``progress``. An optional description of the current
        simulation to use in the ProgressReporter can be passed in as `msg`.
        """
        # Check filename
        filename = os.path.expanduser(filename)

        # Load compression modules
        import zipfile
        try:
            # Ensure zlib is available
            import zlib
            del zlib
        except ImportError:
            raise Exception(
                'This method requires the ``zlib`` module to be installed.')

        # Get size of single and double types on this machine
        dsize = {
            'd': len(array.array('d', [1]).tobytes()),
            'f': len(array.array('f', [1]).tobytes()),
        }

        # Read data
        try:
            f = None
            f = zipfile.ZipFile(filename, 'r')
            # Get ZipInfo objects
            try:
                body = f.getinfo('data.bin')
            except KeyError:
                raise myokit.DataLogReadError('Invalid log file format.')
            try:
                head = f.getinfo('structure.txt')
            except KeyError:
                raise myokit.DataLogReadError('Invalid log file format.')
            # Read file contents
            head = f.read(head).decode(ENC)
            body = f.read(body)
        except zipfile.BadZipfile:
            raise myokit.DataLogReadError('Unable to read log: bad zip file.')
        except zipfile.LargeZipFile:    # pragma: no cover
            raise myokit.DataLogReadError(
                'Unable to read log: zip file requires zip64 support and this'
                ' has not been enabled on this system.')
        finally:
            if f:
                f.close()

        # Create empty log
        log = DataLog()

        # Parse header
        # Number of fields, length of data arrays, data type, time, fields
        head = iter(head.splitlines())
        n = int(next(head))
        data_size = int(next(head))
        data_type = next(head)
        time = next(head)
        if time:
            # Note, this field doesn't have to be present in the log!
            log._time = time
        fields = [x for x in head]
        if len(fields) != n:
            raise myokit.DataLogReadError(
                'Invalid number of fields specified.')

        # Get size of each entry on disk
        if data_size < 0:
            raise myokit.DataLogReadError(
                'Invalid data size: ' + str(data_size) + '.')
        try:
            data_size *= dsize[data_type]
        except KeyError:
            raise myokit.DataLogReadError(
                'Invalid data type: "' + data_type + '".')

        # Parse read data
        fraction = 1.0 / len(fields)
        start, end = 0, 0
        nbody = len(body)
        try:
            if progress:
                progress.enter(msg)
            for k, field in enumerate(fields):
                if progress and not progress.update(k * fraction):
                    return

                # Get new data position
                start = end
                end += data_size
                if end > nbody:
                    raise myokit.DataLogReadError(
                        'Header indicates larger data size than found in body.'
                    )

                # Read data
                ar = array.array(data_type)
                ar.frombytes(body[start:end])
                if sys.byteorder == 'big':  # pragma: no cover
                    ar.byteswap()
                log[field] = ar
        finally:
            if progress:
                progress.exit()
        return log

    @staticmethod
    def load_csv(filename, precision=myokit.DOUBLE_PRECISION):
        """
        Loads a CSV file from disk and returns it as a :class:`DataLog`.

        The CSV file must start with a header line indicating the variable
        names, separated by commas. Each subsequent row should contain the
        values at a single point in time for all logged variables.

        The ``DataLog`` is created using the data type specified by the
        argument ``precision``, regardless of the data type of the stored data.

        The log attempts to set a time variable by searching for a strictly
        increasing variable. In the case of a tie the first strictly increasing
        variable is used. This means logs stored with :meth:`save_csv` can
        safely be read.
        """
        log = DataLog()
        # Check filename
        filename = os.path.expanduser(filename)

        # Typecode dependent on precision
        typecode = 'd' if precision == myokit.DOUBLE_PRECISION else 'f'

        # Error raising function
        def e(line, char, msg):
            raise myokit.DataLogReadError(
                'Syntax error on line ' + str(line) + ', character '
                + str(1 + char) + ': ' + msg)

        quote = '"'
        delim = ','
        with open(filename, 'r', newline=None) as f:
            # Read header
            keys = []   # The log keys, in order of appearance

            # Read first line
            line = f.readline()

            # Ignore comments
            while line.lstrip()[:1] == '#':
                line = f.readline()

            # Stop on EOF (indicated by blank line without line ending)
            if line == '':
                return log

            # Trim end of line
            line = line.rstrip(' \r\n\f;')

            # Get enumerated iterator over characters
            line = enumerate(line)
            try:
                i, c = next(line)
            except StopIteration:
                # Empty line
                return log

            # Whitespace characters to ignore
            whitespace = ' \f\t'

            # Start parsing header fields
            run1 = True
            while run1:
                text = []

                # Skip whitespace
                # (Note: rtrim above and check below mean this will never
                #  raise a StopIteration)
                while c in whitespace:
                    i, c = next(line)

                # Read!
                if c == quote:

                    # Read quoted field + delimiter
                    run2 = True
                    while run2:
                        try:
                            i, c = next(line)
                        except StopIteration:
                            e(1, i, 'Unexpected end-of-line inside quoted'
                                ' string.')

                        if c == quote:
                            try:
                                i, c = next(line)
                                if c == quote:
                                    text.append(quote)
                                elif c == delim or c in whitespace:
                                    run2 = False
                                else:
                                    e(1, i, 'Expecting double quote, delimiter'
                                        ' or end-of-line. Found "' + c + '".')
                            except StopIteration:
                                run1 = run2 = False
                        else:
                            text.append(c)
                else:

                    # Read unquoted field + delimiter
                    while run1 and c != delim:
                        try:
                            text.append(c)
                            i, c = next(line)
                        except StopIteration:
                            run1 = False

                # Append new field to list
                key = ''.join(text)
                if key == '':
                    e(1, i, 'Empty field in header.')
                keys.append(key)

                # Read next character
                try:
                    i, c = next(line)
                except StopIteration:
                    run1 = False

            if c == delim:
                e(1, i, 'Empty field in header.')

            # Create data structure
            m = len(keys)
            lists = []
            for key in keys:
                x = array.array(typecode)
                lists.append(x)
                log[key] = x

            # Read remaining data
            n = 0
            while True:
                row = f.readline()

                # Stop if a blank line is returned: indicates EOF!
                # (Empty line in file still has line ending)
                if row == '':
                    break

                # Strip leading and/or trailing whitespace
                row = row.lstrip().rstrip(' \r\n\f;')

                # Skip blank lines
                if row == '':
                    continue

                # Ignore lines commented with #
                if row[:1] == '#':
                    continue

                # Split row into cells
                row = row.split(delim)
                n += 1
                if len(row) != m:
                    e(
                        n, 0, 'Wrong number of columns found in row '
                        + str(n) + '. Expecting ' + str(m) + ', found '
                        + str(len(row)) + '.')
                try:
                    for k, v in enumerate(row):
                        lists[k].append(float(v))
                except ValueError:
                    e(n, 0, 'Unable to convert found data to floats.')

            # Guess time variable
            for key in keys:
                x = np.array(log[key], copy=False)
                y = x[1:] - x[:-1]
                if np.all(y > 0):
                    log.set_time_key(key)
                    break

            # Return log
            return log

    def npview(self):
        """
        Returns a ``DataLog`` with numpy array views of this log's data.
        """
        log = DataLog()
        log._time = self._time
        for k, v in self.items():
            log[k] = np.asarray(v)
        return log

    def _parse_key(self, key):
        """
        Parses a key used for __getitem__, __setitem__, __delitem__ and
        __contains__.
        """
        if type(key) == tuple:
            name = str(key[0])
            if len(key) == 2 and type(key[1]) not in [int, float]:
                parts = [str(x) for x in key[1]]
            else:
                parts = [str(x) for x in key[1:]]
            parts.append(str(name))
            key = '.'.join(parts)
        return str(key)

    def regularize(self, dt, tmin=None, tmax=None):  # pragma: no cover
        """
        Returns a copy of this DataLog with data points at regularly spaced
        times.

        This method is deprecated and will be removed in future versions of
        Myokit.

        This method requires ``SciPy`` to be installed.

        *Note: While regularize() can be used post-simulation to create fixed
        time-step data from variable time-step data, it is usually better to
        re-run a simulation with fixed time step logging. See*
        :meth:`Simulation.run()` *for details.*

        The first point will be at ``tmin`` if specified or otherwise the first
        time present in the log. All following points will be spaced ``dt``
        time units apart and the final point will be less than or equal to
        ``tmax``. If no value for ``tmax`` is given the final value in the log
        is used.

        This method works by

          1. Finding the indices corresponding to ``tmin`` and ``tmax``.
          2. Creating a spline interpolant with all the data from ``tmin`` to
             ``tmax``. If possible, two points to the left and right of
             ``tmin`` and ``tmax`` will be included in the interpolated data
             set (so *only* if there are at least two values before ``tmin`` or
             two values after ``tmax`` in the data respectively).
          3. Evaluating the interpolant at the regularly spaced points.

        As a result of the cubic spline interpolation, the function may perform
        poorly on large data sets.
        """
        # Deprecated since 2023-06-06
        import warnings
        warnings.warn(
            'The method `regularize` is deprecated and will be removed in'
            ' future versions of Myokit.')

        from scipy.interpolate import UnivariateSpline as Spline
        self.validate()

        # Check time variable
        time = self.time()
        n = len(time)

        # Get left index for splines
        imin = 0
        if tmin is None:
            tmin = time[0]
        elif tmin > time[0]:
            # Find position of tmin in time list, then add two points to the
            # left so that the spline has 4 points
            imin = max(0, np.searchsorted(time, tmin) - 2)

        # Get right index for splines
        imax = n
        if tmax is None:
            tmax = time[-1]
        elif tmax < time[-1]:
            imax = min(n, np.searchsorted(time, tmax) + 2)

        # Get time steps
        steps = 1 + np.floor((tmax - tmin) / dt)
        rtime = tmin + dt * np.arange(0, steps)

        # Create output and return
        out = DataLog()
        out._time = self._time
        out[self._time] = rtime
        time_part = time[imin:imax]
        for key, data in self.items():
            if key != self._time:
                s = Spline(time_part, data[imin:imax], k=1, s=0)
                out[key] = s(rtime)
        return out

    def save(self, filename, precision=myokit.DOUBLE_PRECISION):
        """
        Writes this ``DataLog`` to a binary file.

        The resulting file will be a zip file with the following entries:

        ``header.txt``
            A csv file with the fields ``name, dtype, len`` for each variable.
        ``data.bin``
            The binary data in the order specified by the header.
        ``readme.txt``
            A text file explaining the file format.

        The optional argument ``precision`` allows logs to be stored in single
        precision format, which saves space.
        """
        self.validate()

        # Check filename
        filename = os.path.expanduser(filename)

        # Load compression modules
        import zipfile
        try:
            # Make sure zlib is available
            import zlib
            del zlib
        except ImportError:
            raise Exception(
                'This method requires the `zlib` module to be installed.')

        # Data type
        dtype = 'd' if precision == myokit.DOUBLE_PRECISION else 'f'

        # Create data strings
        head_str = []
        body_str = []   # Will be filled with bytes

        # Number of fields, length of data arrays, data type, time, fields
        head_str.append(str(len(self)))
        head_str.append(str(len(next(iter(self.values())))))
        head_str.append(dtype)

        # Note: the time field might not be present in the log!
        head_str.append(self._time if self._time else '')

        # Write field names and data
        enc = 'utf8'
        for k, v in self.items():
            head_str.append(k)
            # Create array, ensure it's litte-endian
            ar = array.array(dtype, v)
            if sys.byteorder == 'big':  # pragma: no cover
                ar.byteswap()
            body_str.append(ar.tobytes())
        head_str = '\n'.join(head_str)
        body_str = b''.join(body_str)

        # 2018-07-15: Wondering why I chose body-head-readme ordering now...

        # Write
        head = zipfile.ZipInfo('structure.txt')
        head.compress_type = zipfile.ZIP_DEFLATED
        body = zipfile.ZipInfo('data.bin')
        body.compress_type = zipfile.ZIP_DEFLATED
        read = zipfile.ZipInfo('readme.txt')
        read.compress_type = zipfile.ZIP_DEFLATED
        with zipfile.ZipFile(filename, 'w') as f:
            f.writestr(body, body_str)
            f.writestr(head, head_str.encode(enc))
            f.writestr(read, README_SAVE_BIN.encode(enc))

    def save_csv(
            self, filename, precision=myokit.DOUBLE_PRECISION, order=None,
            delimiter=',', header=True):
        """
        Writes this ``DataLog`` to a CSV file, following the syntax
        outlined in RFC 4180 and with a header indicating the field names.

        The resulting file will consist of:

          - A header line containing the names of all logged variables,
            separated by commas. If present, the time variable will be the
            first entry on the line. The remaining keys are ordered using a
            natural sort order.
          - Each following line will be a comma separated list of values in the
            same order as the header line. A line is added for each time point
            logged.

        Arguments:

        ``filename``
            The file to write (existing files will be overwritten without
            warning.
        ``precision``
            If a precision argument (for example ``myokit.DOUBLE_PRECISION``)
            is given, the output will be stored in such a way that this amount
            of precision is guaranteed to be present in the string. If the
            precision argument is set to ``None`` Python's default formatting
            is used, which may lead to smaller files.
        ``order``
            To specify the ordering of the log's arguments, pass in a sequence
            ``order`` with the log's keys.
        ``delimiter``
            This field can be used to set an alternative delimiter. To use
            spaces set ``delimiter=' '``, for tabs: ``delimiter='\\t'``. Note
            that some delimiters (for example '\\n' or '1234') will produce an
            unreadable or invalid csv file.
        ``header``
            Set this to ``False`` to avoid adding a header to the file. Note
            that Myokit will no longer be able to read the written csv file
            without this header.

        *A note about locale settings*: On Windows systems with a locale
        setting that uses the comma as a decimal separator, importing CSV files
        into Microsoft Excel can be troublesome. To correctly import a CSV,
        either (1) Change your locale settings to use "." as a decimal
        separator or (2) Use the import wizard under Data > Get External Data
        to manually specify the correct separator and delimiter.
        """
        self.validate()

        # Check filename
        filename = os.path.expanduser(filename)

        # Set precision
        if precision is None:
            fmat = lambda x: str(x)
        elif precision == myokit.DOUBLE_PRECISION:
            fmat = lambda x: myokit.SFDOUBLE.format(x)
        elif precision == myokit.SINGLE_PRECISION:
            fmat = lambda x: myokit.SFSINGLE.format(x)
        else:
            raise ValueError('Precision level not supported.')

        # Write file
        # EOL: CSV files have DOS line endings by convention. On windows,
        # writing '\n' to a file opened in mode 'w' will actually write '\r\n'
        # which means writing '\r\n' writes '\r\r\n'. To prevent this, open the
        # file in mode 'wb'.
        eol = '\r\n'
        quote = '"'
        escape = '""'
        with open(filename, 'wb') as f:
            # Convert dict structure to ordered sequences
            if order:
                order = [str(x) for x in order]
                if set(order) != set(self.keys()):
                    raise ValueError(
                        'The given `order` sequence must contain all the same'
                        ' keys present in the log.')
                keys = order
                data = [self[x] for x in keys]
            else:
                keys = []
                data = []
                if self._time and self._time in self.keys():
                    # Save time as first variable
                    dat = self[self._time]
                    keys.append(self._time)
                    data.append(dat)
                for key, dat in sorted(
                        self.items(),
                        key=lambda i: myokit.tools.natural_sort_key(i[0])):
                    if key != self._time:
                        keys.append(key)
                        data.append(dat)

            # Number of entries
            m = len(keys)
            if m == 0:
                return

            # Get length of entries
            n = self.length()

            # Write header
            if header:
                line = []
                for key in keys:
                    # Escape quotes within strings
                    line.append(quote + key.replace(quote, escape) + quote)
                f.write((delimiter.join(line) + eol).encode('ascii'))

            # Write data
            data = [iter(x) for x in data]
            for i in range(0, n):
                line = []
                for d in data:
                    line.append(fmat(next(d)))
                f.write((delimiter.join(line) + eol).encode('ascii'))

    def set_time_key(self, key):
        """
        Sets the key under which the time data is stored.
        """
        self._time = None if key is None else str(key)

    def __setitem__(self, key, value):
        return super().__setitem__(
            self._parse_key(key), value)

    def split(self, value):
        """
        Splits the log into a part before and after the time ``value``::

            s = myokit.Simulation(m, p)
            d = s.run(1000)
            d1, d2 = d.split(100)

        In this example, d1 will contain all values up to, but not including,
        t=100. While d2 will contain the values from t=100 and upwards.
        """
        return self.isplit(self.find_after(value))

    def split_periodic(self, period, adjust=False, closed_intervals=True):
        """
        Splits this log into multiple logs, each covering an equal period of
        time. For example a log covering the time span ``[0, 10000]`` can be
        split with period ``1000`` to obtain ten logs covering ``[0, 1000]``,
        ``[1000, 2000]`` etc.

        The split log files can be returned as-is, or with the time variable's
        value adjusted so that all logs appear to cover the same span. To
        enable this option, set ``adjust`` to ``True``.

        By default, the returned intervals are *closed*, so both the left and
        right endpoint are included (if present in the data). This may involve
        the duplication of some data points. To disable this behavior and
        return half-closed endpoints (containing only the left point), set
        ``closed_intervals`` to ``False``.
        """
        # Validate log before starting
        self.validate()

        # Check time variable
        time = self.time()
        if len(time) < 1:
            raise RuntimeError('DataLog entries have zero length.')

        # Check period
        period = float(period)
        if period <= 0:
            raise ValueError('Period must be greater than zero')

        # Get start, end, etc
        tmin = 0    # time[0]
        tmax = time[len(time) - 1]
        nlogs = int(np.ceil((tmax - tmin) / period))

        # No splitting needed? Return clone!
        if nlogs < 2:
            return self.clone()

        # Find split points
        tstarts = tmin + np.arange(nlogs) * period
        istarts = [0] * nlogs
        k = 0
        for i, t in enumerate(time):
            while k < nlogs and t >= tstarts[k]:
                istarts[k] = i
                k += 1

        # Create logs
        logs = []
        for i in range(0, nlogs - 1):
            log = DataLog()
            log._time = self._time

            # Get indices
            imin = istarts[i]
            imax = istarts[i + 1]

            # Include right point endpoint if needed
            if closed_intervals and time[imax] == tstarts[i + 1]:
                imax += 1

            # Select sections of log and append
            for k, v in self.items():
                d = self[k][imin:imax]
                # NumPy? Then copy data
                if isinstance(d, np.ndarray):
                    d = np.array(d, copy=True, dtype=float)
                log[k] = d
            logs.append(log)

        # Last log
        log = DataLog()
        log._time = self._time
        imin = istarts[-1]
        imax = len(time)

        # Not including right endpoints? Then may be required to omit last pt
        if not closed_intervals and time[-1] >= tmin + nlogs * period:
            imax -= 1

        # Select sections of log and append
        for k, v in self.items():
            d = self[k][imin:imax]
            # NumPy? Then copy data
            if isinstance(d, np.ndarray):
                d = np.array(d, copy=True, dtype=float)
            log[k] = d
        logs.append(log)

        # Adjust
        if adjust:
            if isinstance(time, np.ndarray):
                # Fast method for numpy arrays
                for k, log in enumerate(logs):
                    log[self._time] -= k * period
            else:
                for k, log in enumerate(logs):
                    tlist = log[self._time]
                    tdiff = k * period
                    for i in range(len(tlist)):
                        tlist[i] -= tdiff

        # Return
        return logs

    def time(self):
        """
        Returns this log's time array.

        Raises a :class:`myokit.InvalidDataLogError` if the time variable for
        this log has not been specified or an invalid key was given for the
        time variable.
        """
        try:
            return self[self._time]
        except KeyError:
            if self._time is None:
                raise myokit.InvalidDataLogError('No time variable set.')
            else:
                raise myokit.InvalidDataLogError(
                    'Invalid key <' + str(self._time)
                    + '> set for time variable.')

    def time_key(self):
        """
        Returns the name of the time variable stored in this log, or ``None``
        if no time variable was set.
        """
        return self._time

    def trim(self, a, b, adjust=False):
        """
        Returns a copy of this log, with all data before time ``a`` and after
        (and including) time ``b`` removed.

        If ``adjust`` is set to ``True``, all logged times will be lowered by
        ``a``.
        """
        self.validate()
        log = self.itrim(self.find_after(a), self.find_after(b))
        if adjust and self._time in log:
            if isinstance(log[self._time], np.ndarray):
                log[self._time] -= a
            else:
                log[self._time] = [x - a for x in log[self._time]]
        return log

    def trim_left(self, value, adjust=False):
        """
        Returns a copy of this log, with all data before time ``value``
        removed.

        If ``adjust`` is set to ``True``, all logged times will be lowered by
        ``value``.
        """
        self.validate()
        log = self.itrim_left(self.find_after(value))
        if adjust and self._time in log:
            if isinstance(log[self._time], np.ndarray):
                log[self._time] -= value
            else:
                log[self._time] = [x - value for x in log[self._time]]
        return log

    def trim_right(self, value):
        """
        Returns a copy of this log, with all data at times after and including
        ``value`` removed.
        """
        return self.itrim_right(self.find_after(value))

    def validate(self):
        """
        Validates this ``DataLog``. Raises a
        :class:`myokit.InvalidDataLogError` if the log has inconsistencies.
        """
        if self._time:
            if self._time not in self:
                raise myokit.InvalidDataLogError(
                    'Time variable <' + str(self._time)
                    + '> specified but not found in log.')
            dt = np.asarray(self[self._time])
            if np.any(dt[1:] - dt[:-1] < 0):
                raise myokit.InvalidDataLogError(
                    'Time must be non-decreasing.')
        if len(self) > 0:
            n = set([len(v) for v in self.values()])
            if len(n) > 1:
                raise myokit.InvalidDataLogError(
                    'All entries in a data log must have the same length.')

    def variable_info(self):
        """
        Returns a dictionary mapping fully qualified variable names to
        :class:`LoggedvariableInfo` instances, providing information about the
        logged data.

        Comes with the following constraints:

        - Per variable, the data must have a consistent dimensionality. For
          example having a key ``0.membrane.V`` and a key ``1.1.membrane.V``
          would violate this constraint.
        - Per variable, the data must be regular accross dimensions. For
          example if there are ``n`` entries ``0.x.membrane.V``, and there are
          also entries of the form ``1.x.membrane.V`` then the values of ``x``
          must be the same for both cases.

        An example of a dataset that violates the second constraint is::

            0.0.membrane.V
            0.1.membrane.V
            0.2.membrane.V
            1.0.membrane.V
            1.1.membrane.V

        If either of the constraints is violated a ``ValueError`` is raised.
        """
        # The algorithm for condition 2 works by creating a set of the unique
        # entries in each column. The product of the sizes of these sets should
        # equal the total number of entries for a variable.
        # For example:
        #   0 1
        #   0 2     Results in id_sets [(0,1), (1,2,3,4)]
        #   0 3     2 * 4 = 8 != len(id_list)
        #   1 1     So this data must be irregular.
        #   1 2
        #   1 4
        #
        id_lists = {}
        id_sets = {}
        for key in self:
            # Split key into id / name parts
            idx, name = split_key(key)

            # Create tuple version of id
            idx = idx.split('.')
            idx = idx[:-1]
            idx = tuple([int(i) for i in idx])

            # Find or create entry in dict of id lists
            try:
                id_list = id_lists[name]
                id_set = id_sets[name]
            except KeyError:
                # Create entry in id lists dict
                id_lists[name] = id_list = []

                # Create entry in id sets dict (one set for every dimension)
                id_sets[name] = id_set = [set() for x in idx]

            # Check if the dimensions are the same each time a name occurs.
            if id_list and len(id_list[0]) != len(idx):
                key1 = '.'.join([str(x) for x in id_list[0]]) + '.' + name
                key2 = '.'.join([str(x) for x in idx]) + '.' + name
                raise RuntimeError(
                    'Different dimensions used for the same variable. Found: <'
                    + key1 + '> and <' + key2 + '>.')

            # Update the id list
            id_list.append(idx)

            # Update the id set
            for k, i in enumerate(idx):
                id_set[k].add(i)

        # Create variable info objects
        infos = {}
        for name, id_list in id_lists.items():
            id_set = id_sets[name]

            # Check if the data is regular.
            n = len(id_list)
            m = 1
            for x in id_set:
                m *= len(x)

            if n != m:
                raise RuntimeError(
                    'Irregular data used for variable <' + str(name) + '>')

            # Create variable info object
            infos[name] = info = LoggedVariableInfo()
            info._name = name
            info._dimension = len(id_set)
            info._size = tuple([len(x) for x in id_set])

            # Add sorted ids
            if id_list[0]:
                id_list.sort()
            info._ids = id_list

            # Add sorted keys
            s = '.' + name if id_list[0] else name
            info._keys = ['.'.join([str(x) for x in y]) + s for y in id_list]

        return infos


class LoggedVariableInfo:
    """
    Contains information about the log entries for each variable. These objects
    should only be created by :meth:`DataLog.variable_info()`.
    """
    def __init__(self):
        self._dimension = None
        self._ids = None
        self._keys = None
        self._size = None
        self._name = None

    def dimension(self):
        """
        Returns the dimensions of the logged data for this variable, as an
        integer.
        """
        return self._dimension

    def ids(self):
        """
        Returns an iterator over all available ids for this variable, such
        that the second index (y in the simulation) changes fastest. For
        example, for log entries::

            0.0.membrane.V
            0.1.membrane.V
            0.2.membrane.V
            1.0.membrane.V
            1.1.membrane.V
            1.2.membrane.V

        the returned result would iterate over::

            [(0, 0), (0, 1), (0, 2), (1, 0), (1, 1), (1, 2)]

        The keys are returned in the same order as the ids.
        """
        return iter(self._ids)

    def is_regular_grid(self):
        """
        Returns True if the following conditions are met:

        - The data 2 dimensional
        - The data is continuous: along each dimension the first data point is
          indexed as ``0`` and the last as ``Ni-1``, where ``Ni`` is the size
          in that dimension.

        """
        nx, ny = self._size
        return (
            self._dimension == 2
            and self._ids[0][0] == 0
            and self._ids[0][1] == 0
            and self._ids[-1][0] == nx - 1
            and self._ids[-1][1] == ny - 1
        )

    def keys(self):
        """
        Returns an iterator over all available keys for this variable, such
        that the second index (y in the simulation) changes fastest. For
        example, for log entries::

            0.0.membrane.V
            1.0.membrane.V
            0.1.membrane.V
            1.1.membrane.V
            0.2.membrane.V
            1.2.membrane.V

        the returned iterator would produce ``"0.0.membrane.V"``, then
        ``"0.1.membrane.V"`` etc.

        The ids are returned in the same order as the keys.
        """
        return iter(self._keys)

    def size(self):
        """
        Returns a tuple containing the size i.e. the number of entries for
        the corresponding variable in each dimension.

        For example, with the following log entries for `membrane.V`::

            0.membrane.V
            1.membrane.V
            2.membrane.V

        the corresponding size would be ``(3)``.

        A size of ``3`` doesn't guarantee the final entry is for cell number
        ``2``. For example::

            0.membrane.V
            10.membrane.V
            20.membrane.V

        would also return size ``(3)``

        In higher dimensions::

            0.0.membrane.V
            0.1.membrane.V
            0.2.membrane.V
            1.0.membrane.V
            1.1.membrane.V
            1.2.membrane.V

        This would return ``(2,3)``.

        Similarly, in a single cell scenario or for global variables, for
        exmaple::

            engine.time

        Would have size ``()``.
        """
        return self._size

    def name(self):
        """
        Returns the variable name.
        """
        return self._name

    def to_long_string(self):   # pragma: no cover (debug function)
        out = [self._name]
        out.append('  dimension: ' + str(self._dimension))
        out.append('  size: ' + ', '.join([str(x) for x in self._size]))
        out.append('  keys:')
        out.extend(['    ' + x for x in self._keys])
        return '\n'.join(out)


def prepare_log(
        log, model, dims=None, global_vars=None, if_empty=myokit.LOG_NONE,
        allowed_classes=myokit.LOG_ALL, precision=myokit.DOUBLE_PRECISION):
    """
    Returns a :class:`DataLog` for simulation classes based on a ``log``
    argument passed in by the user. The model the simulations will be based on
    should be passed in as ``model``.

    The ``log`` argument can take on one of four forms:

    An existing simulation log
        In this case, the log is tested for compatibility with the given model
        and simulation dimensions. For single-cell simulations all keys in the
        log must correspond to the qname of a loggable variable (IE not a
        constant). For multi-cellular simulations this means all keys in the
        log must have the form "x.component.variable" where "x" is the cell
        index (for example "1" or "0.3").
    A list (or other sequence) of variable names to log.
        In this case, the list is converted to a DataLog object. All
        arguments in the list must be either strings corresponding to the
        variables' qnames (so "membrane.V") or variable objects from the given
        model.
        In multi-cell scenarios, passing in the qname of a variable (for
        example "membrane.V") will cause every cell's instance of this variable
        to be logged. To log only specific cells' values, pass in the indexed
        name (for example "1.2.membrane.V").
    An integer flag
        One of the following integer flags:

        ``myokit.LOG_NONE``
            Don't log any variables.
        ``myokit.LOG_STATE``
            Log all state variables.
        ``myokit.LOG_BOUND``
            Log all variables bound to an external value. The method will
            assume any bound variables still present in the model will be
            provided by the simulation engine.
        ``myokit.LOG_INTER``
            Log all intermediary variables.
        ``myokit.LOG_DERIV``
            Log the derivatives of the state variables.
        ``myokit.LOG_ALL``
            Combines all the previous flags.

        Flags can be chained together, for example
        ``log=myokit.LOG_STATE+myokit.LOG_BOUND`` will log all bound variables
        and all states.
    ``None``
        In this case the value from ``if_empty`` will be copied into log before
        the function proceeds to build a log.

    For multi-dimensional logs the simulation dimensions can be passed in as a
    tuple of dimension sizes, for example (10,) for a cable of 10 cells and
    (30,20) for a 30 by 20 piece of tissue.

    Simulations can define variables to be either `per-cell` or `global`. Time,
    for example, is typically a global variable while membrane potential will
    be stored per cell. To indicate which is which, a list of global variables
    can be passed in as ``global_vars``.

    The argument ``if_empty`` is used to set a default argument if ``log`` is
    is given as ``None``.

    The argument ``allowed_classes`` is an integer flag that determines which
    type of variables are allowed in this log.

    When a new DataLog is created by this method, the internal storage
    uses arrays from the array module. The data type for these new arrays can
    be specified using the ``precision`` argument.
    """
    # Typecode dependent on precision
    typecode = 'd' if precision == myokit.DOUBLE_PRECISION else 'f'

    # Get all options for dimensionality
    if dims is None:
        dims = ()
    ndims = len(dims)
    if ndims == 0:
        dcombos = ['']
    else:
        dcombos = ['.'.join([str(y) for y in x]) + '.' for x in _dimco(*dims)]

    # Check given list of global variables
    if global_vars is None:
        global_vars = []
    else:
        for var in global_vars:
            try:
                v = model.get(var)
            except KeyError:
                raise ValueError(
                    'Unknown variable specified in global_vars <'
                    + str(var) + '>.')
            if v.is_state():
                raise ValueError('State cannot be global variable.')

    # Function to check if variable is allowed (doesn't handle derivatives)
    def check_if_allowed_class(var):
        if var.is_constant():
            raise ValueError(
                'This log does not support constants, got <'
                + str(var) + '>.')
        elif var.is_state():
            if not myokit.LOG_STATE & allowed_classes:
                raise ValueError(
                    'This log does not support state variables, got <'
                    + str(var) + '>.')
        elif var.is_bound():
            if not myokit.LOG_BOUND & allowed_classes:
                raise ValueError(
                    'This log does not support bound variables, got <'
                    + str(var) + '>.')
        elif not myokit.LOG_INTER & allowed_classes:
            raise ValueError(
                'This log does not support intermediary variables, got <'
                + str(var) + '>.')

    #
    # First option, no log argument given, use the "if_empty" option
    #

    if log is None:
        # Check if if_empty matches allowed_classes
        # (AKA test if ``if_empty`` is contained in ``allowed_classes``)
        if if_empty & allowed_classes == if_empty:
            log = if_empty
        else:
            # This one's only for programmers :-)
            raise ValueError(
                'The option given for `if_empty` should be an allowed class.')

    #
    # Second option, log given as integer flag: create a simulation log and
    # return it.
    #

    if type(log) == int:

        # Log argument given as flag
        flag = log
        log = myokit.DataLog()
        if flag == myokit.LOG_ALL:
            flag = allowed_classes

        if myokit.LOG_STATE & flag:

            # Check if allowed
            if not (myokit.LOG_STATE & allowed_classes):
                raise ValueError('This log does not support state variables.')

            # Add states
            for s in model.states():
                name = s.qname()
                for c in dcombos:
                    log[c + name] = array.array(typecode)
            flag -= myokit.LOG_STATE

        if myokit.LOG_BOUND & flag:

            # Check if allowed
            if not (myokit.LOG_BOUND & allowed_classes):
                raise ValueError('This log does not support bound variables.')

            # Add bound variables
            for label, var in model.bindings():
                name = var.qname()
                if name in global_vars:
                    log[name] = array.array(typecode)
                else:
                    for c in dcombos:
                        log[c + name] = array.array(typecode)
            flag -= myokit.LOG_BOUND

        if myokit.LOG_INTER & flag:

            # Check if allowed
            if not (myokit.LOG_INTER & allowed_classes):
                raise ValueError(
                    'This log does not support intermediary variables.')

            # Add intermediary variables
            for var in model.variables(inter=True, deep=True):
                name = var.qname()
                if name in global_vars:
                    log[name] = array.array(typecode)
                else:
                    for c in dcombos:
                        log[c + name] = array.array(typecode)
            flag -= myokit.LOG_INTER

        if myokit.LOG_DERIV & flag:

            # Check if allowed
            if not (myokit.LOG_DERIV & allowed_classes):
                raise ValueError('This log does not support time-derivatives.')

            # Add state derivatives
            for var in model.states():
                name = var.qname()
                for c in dcombos:
                    log['dot(' + c + name + ')'] = array.array(typecode)
            flag -= myokit.LOG_DERIV

        if flag != 0:
            raise ValueError('One or more unknown flags given as log.')

        # Set time variable
        time = model.time().qname()
        if time in log:
            log.set_time_key(time)

        # Return
        return log

    #
    # Third option, a dict or DataLog is given. Test if it's suitable for this
    # simulation.
    #

    if isinstance(log, dict):

        # Ensure it's a DataLog
        if not isinstance(log, myokit.DataLog):
            log = myokit.DataLog(log)

        # Set time variable
        time = model.time().qname()
        if time in log:
            log.set_time_key(time)

        # Ensure the log is valid
        log.validate()

        # Check dict keys
        keys = set(log.keys())
        if len(keys) == 0:
            return log

        for key in keys:
            # Handle derivatives
            deriv = key[0:4] == 'dot(' and key[-1:] == ')'
            if deriv:
                key = key[4:-1]

            # Split of index / name
            kdims, kname = split_key(key)

            # Test name-key
            try:
                var = model.get(kname)
            except KeyError:
                raise ValueError(
                    'Unknown variable <' + str(kname) + '> in log.')

            # Check if in class of allowed variables
            if deriv:
                if not myokit.LOG_DERIV & allowed_classes:
                    raise ValueError(
                        'This log does not support derivatives, got <'
                        + key + '>.')
                if not var.is_state():
                    raise ValueError(
                        'Cannot log time derivative of non-state <'
                        + var.qname() + '>.')
            else:
                check_if_allowed_class(var)

            # Check dimensions
            if kdims:

                # Raise error if global
                if kname in global_vars:
                    raise ValueError(
                        'Cannot specify a cell index for global logging'
                        ' variable <' + str(kname) + '>.')

                # Test dim key
                if kdims not in dcombos:
                    raise ValueError(
                        'Invalid index <' + str(kdims) + '> in log.')

            elif dims:

                # Raise error if non-global variable is used in multi-cell log
                if kname not in global_vars:
                    raise ValueError(
                        'DataLog contains non-indexed entry for'
                        ' cell-specific variable <' + str(kname) + '>.')

        # Check dict values can be appended to
        m = 'append'
        for v in log.values():
            if not (hasattr(v, m) and callable(getattr(v, m))):
                raise ValueError(
                    'Logging dict must map qnames to objects'
                    ' that support the append() method.')

        # Return
        return log

    #
    # Fourth option, a sequence of variable names, either global or local.
    #

    # Check if list interface works
    # If not, then raise exception
    try:
        if len(log) > 0:
            log[0]
    except Exception:
        raise ValueError(
            'Argument `log` has unexpected type. Expecting None, integer flag,'
            ' sequence of names, dict or DataLog.')

    if isinstance(log, str):
        raise ValueError(
            'String passed in as `log` argument, should be list'
            ' or other sequence containing strings.')

    # Parse log argument as list
    lst = log
    log = myokit.DataLog()
    checked_knames = set()
    for key in lst:

        # Allow variable objects and LhsExpressions
        if isinstance(key, myokit.Variable):
            key = key.qname()
        elif isinstance(key, myokit.LhsExpression):
            key = str(key)

        # Handle derivatives
        deriv = key[0:4] == 'dot(' and key[-1:] == ')'
        if deriv:
            key = key[4:-1]

        # Split off cell indexes
        kdims, kname = split_key(key)

        # Don't re-test multi-dim vars
        if kname not in checked_knames:

            # Test if name key points to valid variable
            try:
                var = model.get(kname)
            except KeyError:
                raise ValueError(
                    'Unknown variable <' + str(kname) + '> in list.')

            # Check if in class of allowed variables
            if deriv:
                if not myokit.LOG_DERIV & allowed_classes:
                    raise ValueError(
                        'This log does not support derivatives, got <'
                        + key + '>.')

                if not var.is_state():
                    raise ValueError(
                        'Cannot log time derivative of non-state <'
                        + var.qname() + '>.')
            else:
                check_if_allowed_class(var)
            checked_knames.add(kname)

        # Add key to log
        if kdims:

            # Raise error if global
            if kname in global_vars:
                raise ValueError(
                    'Cannot specify a cell index for global logging variable'
                    ' <' + str(kname) + '>.')

            # Test dim key
            if kdims not in dcombos:
                raise ValueError(
                    'Invalid index <' + str(kdims) + '> in list.')

            key = kdims + kname if not deriv else 'dot(' + kdims + kname + ')'
            log[key] = array.array(typecode)

        else:

            if kname in global_vars:
                key = kname if not deriv else 'dot(' + kname + ')'
                log[key] = array.array(typecode)
            else:
                for c in dcombos:
                    key = c + kname if not deriv else 'dot(' + c + kname + ')'
                    log[key] = array.array(typecode)

    # Set time variable
    time = model.time().qname()
    if time in log:
        log.set_time_key(time)

    # Return
    return log


def _dimco(*dims):
    """
    Generates all the combinations of a certain set of integer dimensions. For
    example given ``dims=(2, 3)`` it returns::

        (0, 0), (1, 0), (0, 1), (1, 1), (0, 2), (1, 2)

    """
    n = len(dims) - 1
    """
    def inner(dims, index, prefix):
        if index == n:
            for i in range(0,dims[index]):
                yield prefix + (i,)
        else:
            for i in range(0,dims[index]):
                prefix2 = prefix + (i, )
                for y in inner(dims, index + 1, prefix2):
                    yield y
    return inner(dims, 0, ())
    """
    def inner(dims, index, postfix):
        if index == 0:
            for i in range(0, dims[index]):
                yield (i,) + postfix
        else:
            for i in range(0, dims[index]):
                postfix2 = (i, ) + postfix
                for y in inner(dims, index - 1, postfix2):
                    yield y
    return inner(dims, n, ())


def split_key(key):
    """
    Splits a log entry name into a cell index part and a variable name part.

    The cell index will be an empty string for 0d entries or global variables.
    For higher dimensional cases it will be the cell index in each dimension,
    followed by a period, for example: ``15.2.``.

    The two parts returned by split_key may always be concatenated to obtain
    the original entry.
    """
    m = ID_NAME_PATTERN.match(key, 0)
    if m:
        return key[:m.end()], key[m.end():]
    else:
        return '', key



#!/usr/bin/env python3
#
# Tests Myokit's SBML api.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import unittest

import myokit
import myokit.formats
import myokit.formats.sbml as sbml

from myokit.formats.sbml._api import _MyokitConverter as X

from myokit.tests import WarningCollector


class TestCompartment(unittest.TestCase):
    """
    Unit tests for :class:`Compartment`.
    """

    @classmethod
    def setUpClass(cls):
        cls.model = sbml.Model(name='model')
        cls.sid = 'compartment'
        cls.c = cls.model.add_compartment(sid=cls.sid)

    def test_bad_model(self):

        model = 'model'
        sid = 'compartment'
        self.assertRaisesRegex(
            sbml.SBMLError, '<', sbml.Compartment, model, sid)

    def test_initial_value(self):

        # Check default initial value
        self.assertIsNone(self.c.initial_value())

        # Check bad value
        expr = 2
        self.assertRaisesRegex(
            sbml.SBMLError, '<', self.c.set_initial_value, expr)

        # Check good value
        expr = myokit.Number(2)
        self.c.set_initial_value(expr)

        self.assertEqual(self.c.initial_value(), expr)

    def test_is_rate(self):

        # Check default
        self.assertFalse(self.c.is_rate())

        # Check setting rate to true
        expr = myokit.Number(2)
        self.c.set_value(value=expr, is_rate=True)

        self.assertTrue(self.c.is_rate())

    def test_sid(self):
        self.assertEqual(self.c.sid(), self.sid)

    def test_size_units(self):

        # Check default units
        unit = myokit.units.dimensionless
        self.assertEqual(self.c.size_units(), unit)

        # Check spatial dimensions = 1
        self.c.set_spatial_dimensions(dimensions=1)
        unit = myokit.units.meter
        self.model.set_length_units(unit)

        self.assertEqual(self.c.size_units(), unit)

        # Check spatial dimensions = 2
        self.c.set_spatial_dimensions(dimensions=2)
        unit = myokit.units.meter**2
        self.model.set_area_units(unit)

        self.assertEqual(self.c.size_units(), unit)

        # Check spatial dimensions = 3
        self.c.set_spatial_dimensions(dimensions=3)
        unit = myokit.units.L
        self.model.set_volume_units(unit)

        self.assertEqual(self.c.size_units(), unit)

        # Check bad size units
        unit = 'mL'
        self.assertRaisesRegex(
            sbml.SBMLError, '<', self.c.set_size_units, unit)

        # Check valid size units
        unit = myokit.units.L * 1E-3
        self.c.set_size_units(unit)

        self.assertEqual(self.c.size_units(), unit)

    def test_spatial_dimensions(self):

        # Test valid dimensions
        dim = 2.31
        self.c.set_spatial_dimensions(dim)

        self.assertEqual(self.c.spatial_dimensions(), dim)

    def test_string_representation(self):

        self.assertEqual(str(self.c), '<Compartment ' + self.sid + '>')

    def test_value(self):

        # Check bad value
        expr = 2
        self.assertRaisesRegex(
            sbml.SBMLError, '<', self.c.set_value, expr)

        # Check good value
        expr = myokit.Number(2)
        self.c.set_value(expr)

        self.assertEqual(self.c.value(), expr)


class TestModel(unittest.TestCase):
    """
    Unit tests for :class:`Model`.
    """

    def test_area_units(self):

        model = sbml.Model(name='model')

        area_units = myokit.units.meter ** 2
        model.set_area_units(area_units)

        self.assertEqual(model.area_units(), area_units)

    def test_assignable(self):
        # Tests assignable() and assignable_or_csymbol()

        model = sbml.Model(name='model')

        # Add assignables to model
        c_sid = 'compartment'
        c = model.add_compartment(c_sid)

        p_sid = 'parameters'
        p = model.add_parameter(p_sid)

        s_sid = 'species'
        comp = sbml.Compartment(model=model, sid=c_sid)
        s = model.add_species(compartment=comp, sid=s_sid)

        # Check that all assignables are accessible
        t = 'http://www.sbml.org/sbml/symbols/time'
        self.assertEqual(model.assignable(c_sid), c)
        self.assertEqual(model.assignable(p_sid), p)
        self.assertEqual(model.assignable(s_sid), s)
        self.assertRaises(KeyError, model.assignable, t)

        # Check that assignables and csymbols can be obtained with
        # assignable_or_csymbol:
        self.assertEqual(model.assignable_or_csymbol(c_sid), c)
        self.assertEqual(model.assignable_or_csymbol(p_sid), p)
        self.assertEqual(model.assignable_or_csymbol(s_sid), s)
        self.assertEqual(model.assignable_or_csymbol(t), model.time())
        self.assertRaises(
            KeyError, model.assignable,
            'http://www.sbml.org/sbml/symbols/beer')
        self.assertRaises(
            KeyError, model.assignable, 'http://google.com')
        self.assertRaises(
            KeyError, model.assignable, 'https://google.com')

    def test_base_unit(self):

        model = sbml.Model(name='model')

        # Check all base units
        self.assertEqual(
            model.base_unit('ampere'), myokit.units.A)
        self.assertEqual(
            model.base_unit('avogadro'),
            myokit.parse_unit('1 (6.02214179e23)'))
        self.assertEqual(model.base_unit('becquerel'), myokit.units.Bq)
        self.assertEqual(model.base_unit('candela'), myokit.units.cd)
        self.assertEqual(model.base_unit('coulomb'), myokit.units.C)
        self.assertEqual(
            model.base_unit('dimensionless'), myokit.units.dimensionless)
        self.assertEqual(model.base_unit('farad'), myokit.units.F)
        self.assertEqual(model.base_unit('gram'), myokit.units.g)
        self.assertEqual(model.base_unit('gray'), myokit.units.Gy)
        self.assertEqual(model.base_unit('henry'), myokit.units.H)
        self.assertEqual(model.base_unit('hertz'), myokit.units.Hz)
        self.assertEqual(model.base_unit('item'), myokit.units.dimensionless)
        self.assertEqual(model.base_unit('joule'), myokit.units.J)
        self.assertEqual(model.base_unit('katal'), myokit.units.kat)
        self.assertEqual(model.base_unit('kelvin'), myokit.units.K)
        self.assertEqual(model.base_unit('kilogram'), myokit.units.kg)
        self.assertEqual(model.base_unit('liter'), myokit.units.L)
        self.assertEqual(model.base_unit('litre'), myokit.units.L)
        self.assertEqual(model.base_unit('lumen'), myokit.units.lm)
        self.assertEqual(model.base_unit('lux'), myokit.units.lux)
        self.assertEqual(model.base_unit('meter'), myokit.units.m)
        self.assertEqual(model.base_unit('metre'), myokit.units.m)
        self.assertEqual(model.base_unit('mole'), myokit.units.mol)
        self.assertEqual(model.base_unit('newton'), myokit.units.N)
        self.assertEqual(model.base_unit('ohm'), myokit.units.ohm)
        self.assertEqual(model.base_unit('pascal'), myokit.units.Pa)
        self.assertEqual(model.base_unit('radian'), myokit.units.rad)
        self.assertEqual(model.base_unit('second'), myokit.units.s)
        self.assertEqual(model.base_unit('siemens'), myokit.units.S)
        self.assertEqual(model.base_unit('sievert'), myokit.units.Sv)
        self.assertEqual(model.base_unit('steradian'), myokit.units.sr)
        self.assertEqual(model.base_unit('tesla'), myokit.units.T)
        self.assertEqual(model.base_unit('volt'), myokit.units.V)
        self.assertEqual(model.base_unit('watt'), myokit.units.W)
        self.assertEqual(model.base_unit('weber'), myokit.units.Wb)

        # Check celsius (not supported)
        self.assertRaisesRegex(
            sbml.SBMLError,
            'The units "celsius" are not supported.',
            model.base_unit,
            'celsius')

        # Check bad base unit
        self.assertRaisesRegex(
            sbml.SBMLError, '<', model.base_unit, 'some unit')

    def test_compartment(self):

        model = sbml.Model(name='model')

        # Test bad sid
        sid = ';'
        self.assertRaisesRegex(
            sbml.SBMLError, 'Invalid SId "', model.add_compartment, sid)

        # Test good sid
        sid = 'compartment'
        model.add_compartment(sid)

        self.assertIsInstance(model.compartment(sid), sbml.Compartment)

        # Test duplicate sid
        sid = 'compartment'
        self.assertRaisesRegex(
            sbml.SBMLError, 'Duplicate SId "', model.add_compartment, sid)

    def test_conversion_factor(self):

        model = sbml.Model(name='model')

        # Test default value
        self.assertIsNone(model.conversion_factor())

        # Bad conversion factor
        self.assertRaisesRegex(
            sbml.SBMLError, '<', model.set_conversion_factor, 10)

        # Good conversion factor
        factor = sbml.Parameter(model, 'parameter')
        model.set_conversion_factor(factor)

        self.assertIsInstance(
            model.conversion_factor(), sbml.Parameter)

    def test_extent_units(self):

        model = sbml.Model(name='model')

        # Test default
        self.assertEqual(model.extent_units(), myokit.units.dimensionless)

        # Test bad extent units
        self.assertRaisesRegex(
            sbml.SBMLError, '<', model.set_extent_units, 'mg')

        # Test good extent units
        model.set_extent_units(myokit.units.g)

        self.assertEqual(model.extent_units(), myokit.units.g)

    def test_length_units(self):

        model = sbml.Model(name='model')

        # Test default
        self.assertEqual(model.length_units(), myokit.units.dimensionless)

        # Test bad length units
        self.assertRaisesRegex(
            sbml.SBMLError, '<', model.set_length_units, 'm')

        # Test good length units
        model.set_length_units(myokit.units.meter)

        self.assertEqual(model.length_units(), myokit.units.meter)

    def test_name(self):

        name = 'model'
        model = sbml.Model(name=name)

        self.assertEqual(model.name(), name)

    def test_notes(self):

        model = sbml.Model(name='model')

        notes = 'Here are some notes.'
        model.set_notes(notes)

        self.assertEqual(model.notes(), notes)

    def test_parameter(self):

        model = sbml.Model(name='model')

        # Test bad sid
        sid = ';'
        self.assertRaisesRegex(
            sbml.SBMLError, 'Invalid SId "', model.add_parameter, sid)

        # Test good sid
        sid = 'parameter'
        model.add_parameter(sid)

        self.assertIsInstance(model.parameter(sid), sbml.Parameter)

        # Test duplicate sid
        sid = 'parameter'
        self.assertRaisesRegex(
            sbml.SBMLError, 'Duplicate SId "', model.add_parameter, sid)

    def test_reaction(self):

        model = sbml.Model(name='model')

        # Test bad sid
        sid = ';'
        self.assertRaisesRegex(
            sbml.SBMLError, 'Invalid SId "', model.add_reaction, sid)

        # Test good sid
        sid = 'reaction'
        model.add_reaction(sid)

        self.assertIsInstance(model.reaction(sid), sbml.Reaction)

        # Test duplicate sid
        sid = 'reaction'
        self.assertRaisesRegex(
            sbml.SBMLError, 'Duplicate SId "', model.add_reaction, sid)

    def test_species(self):

        model = sbml.Model(name='model')

        # Test bad compartment
        comp = 'some compartment'
        sid = 'species'
        self.assertRaisesRegex(
            sbml.SBMLError, '<', model.add_species, comp, sid)

        # Test bad sid
        c_sid = 'compartment'
        comp = sbml.Compartment(model=model, sid=c_sid)
        sid = ';'
        self.assertRaisesRegex(
            sbml.SBMLError, 'Invalid SId "', model.add_species, comp, sid)

        # Test good sid
        c_sid = 'compartment'
        comp = sbml.Compartment(model=model, sid=c_sid)
        sid = 'species'

        model.add_species(compartment=comp, sid=sid)

        self.assertIsInstance(model.species(sid), sbml.Species)

        # Test duplicate sid
        sid = 'species'
        self.assertRaisesRegex(
            sbml.SBMLError, 'Duplicate SId "', model.add_species, comp, sid)

    def test_string_representation(self):

        # Check no name provided
        model = sbml.Model(name=None)

        self.assertEqual(str(model), '<SBMLModel>')

        # Check name provided
        name = 'model'
        model = sbml.Model(name='model')

        self.assertEqual(str(model), '<SBMLModel ' + name + '>')

    def test_substance_units(self):

        model = sbml.Model(name='model')

        # Test default
        self.assertEqual(model.substance_units(), myokit.units.dimensionless)

        # Test bad substance units
        self.assertRaisesRegex(
            sbml.SBMLError, '<', model.set_substance_units, 'mg')

        # Test good substance units
        model.set_substance_units(myokit.units.g)

        self.assertEqual(model.substance_units(), myokit.units.g)

    def test_time_symbol(self):

        model = sbml.Model(name='model')
        time = model.time()
        turl = 'http://www.sbml.org/sbml/symbols/time'
        self.assertIsInstance(time, sbml.CSymbolVariable)
        self.assertEqual(str(time), '<CSymbolVariable ' + turl + '>')
        self.assertEqual(time.definition_url(), turl)

    def test_time_units(self):

        model = sbml.Model(name='model')

        # Test default
        self.assertEqual(model.time_units(), myokit.units.dimensionless)

        # Test bad time units
        self.assertRaisesRegex(
            sbml.SBMLError, '<', model.set_time_units, 's')

        # Test good time units
        model.set_time_units(myokit.units.s)

        self.assertEqual(model.time_units(), myokit.units.s)

    def test_unit(self):

        model = sbml.Model(name='model')

        # Check bad units
        unitsid = 'some_unit'
        self.assertRaisesRegex(
            sbml.SBMLError, 'The unit SID <', model.unit, unitsid)

        # Check base unit
        unitsid = 'ampere'
        self.assertEqual(model.unit(unitsid), myokit.units.A)

        # Check invalid unitsid for user-defined unit
        unitsid = ';'
        unit = myokit.units.dimensionless
        self.assertRaisesRegex(
            sbml.SBMLError, 'Invalid UnitSId "', model.add_unit, unitsid, unit)

        # Check unitsid for user-defined unit that coincides with base unit
        unitsid = 'ampere'
        unit = myokit.units.dimensionless
        self.assertRaisesRegex(
            sbml.SBMLError,
            'User unit overrides built-in unit: "',
            model.add_unit,
            unitsid,
            unit)

        # Check invalid user-defined unit
        unitsid = 'some_unit'
        unit = 'ampere'
        self.assertRaisesRegex(
            sbml.SBMLError,
            'Unit "',
            model.add_unit,
            unitsid,
            unit)

        # Check good user-defined unit
        unitsid = 'some_unit'
        unit = myokit.units.A
        model.add_unit(unitsid, unit)

        self.assertEqual(model.unit(unitsid), unit)

        # Check duplicate unitsid
        unitsid = 'some_unit'
        unit = myokit.units.A
        self.assertRaisesRegex(
            sbml.SBMLError, 'Duplicate UnitSId: "',
            model.add_unit, unitsid, unit)

    def test_volume_units(self):

        model = sbml.Model(name='model')

        # Test default
        self.assertEqual(model.volume_units(), myokit.units.dimensionless)

        # Test bad volume units
        self.assertRaisesRegex(
            sbml.SBMLError, '<', model.set_volume_units, 'L')

        # Test good volume units
        model.set_volume_units(myokit.units.L)

        self.assertEqual(model.volume_units(), myokit.units.L)


class TestParameter(unittest.TestCase):
    """
    Unit tests for :class:`Parameter`.
    """

    @classmethod
    def setUpClass(cls):
        cls.model = sbml.Model(name='model')
        cls.sid = 'parameter'
        cls.p = cls.model.add_parameter(sid=cls.sid)

    def test_bad_model(self):

        model = 'model'
        sid = 'parameter'
        self.assertRaisesRegex(
            sbml.SBMLError, '<', sbml.Parameter, model, sid)

    def test_initial_value(self):

        # Check default initial value
        self.assertIsNone(self.p.initial_value())

        # Check bad value
        expr = 2
        self.assertRaisesRegex(
            sbml.SBMLError, '<', self.p.set_initial_value, expr)

        # Check good value
        expr = myokit.Number(2)
        self.p.set_initial_value(expr)

        self.assertEqual(self.p.initial_value(), expr)

    def test_is_rate(self):

        # Check default
        self.assertFalse(self.p.is_rate())

        # Check setting rate to true
        expr = myokit.Number(2)
        self.p.set_value(value=expr, is_rate=True)

        self.assertTrue(self.p.is_rate())

    def test_sid(self):
        self.assertEqual(self.p.sid(), self.sid)

    def test_string_representation(self):
        self.assertEqual(str(self.p), '<Parameter ' + self.sid + '>')

    def test_units(self):

        # Check default units
        self.assertIsNone(self.p.units())

        # Check bad size units
        unit = 'mL'
        self.assertRaisesRegex(
            sbml.SBMLError, '<', self.p.set_units, unit)

        # Check valid size units
        unit = myokit.units.L * 1E-3
        self.p.set_units(unit)

        self.assertEqual(self.p.units(), unit)

    def test_value(self):

        # Check bad value
        expr = 2
        self.assertRaisesRegex(
            sbml.SBMLError, '<', self.p.set_value, expr)

        # Check good value
        expr = myokit.Number(2)
        self.p.set_value(expr)

        self.assertEqual(self.p.value(), expr)


class TestReaction(unittest.TestCase):
    """
    Unit tests for :class:`Reaction`.
    """

    @classmethod
    def setUpClass(cls):
        cls.model = sbml.Model(name='model')
        cls.sid = 'reaction'
        cls.r = cls.model.add_reaction(sid=cls.sid)

    def test_bad_model(self):

        model = 'model'
        sid = 'reaction'
        self.assertRaisesRegex(
            sbml.SBMLError, '<', sbml.Reaction, model, sid)

    def test_kinetic_law(self):

        # Check default
        self.assertIsNone(self.r.kinetic_law())

        # Check bad kinetic law
        expr = '2 * s'
        self.assertRaisesRegex(
            sbml.SBMLError, '<', self.r.set_kinetic_law, expr)

        # Check good kinetic law
        expr = myokit.Multiply(myokit.Number(2), myokit.Name('s'))
        self.r.set_kinetic_law(expr)

        self.assertEqual(self.r.kinetic_law(), expr)

    def test_modifiers(self):

        # Check bad species
        sid = 'modifier'
        species = 'species'
        self.assertRaisesRegex(
            sbml.SBMLError, '<', self.r.add_modifier, species, sid)

        # Check invalid sid
        sid = ';'
        compartment = sbml.Compartment(self.model, sid='compartment')
        species = sbml.Species(
            compartment=compartment,
            sid='species',
            is_amount=False,
            is_constant=False,
            is_boundary=False)
        self.assertRaisesRegex(
            sbml.SBMLError, 'Invalid SId "', self.r.add_modifier, species, sid)

        # Check good sid
        sid = 'modifier'
        compartment = sbml.Compartment(self.model, sid='compartment')
        species = sbml.Species(
            compartment=compartment,
            sid='species',
            is_amount=False,
            is_constant=False,
            is_boundary=False)
        self.r.add_modifier(species, sid)

        self.assertEqual(len(self.r.modifiers()), 1)
        self.assertIsInstance(
            self.r.modifiers()[0], sbml.ModifierSpeciesReference)

        # Check duplicate sid
        sid = 'modifier'
        compartment = sbml.Compartment(self.model, sid='compartment')
        species = sbml.Species(
            compartment=compartment,
            sid='species',
            is_amount=False,
            is_constant=False,
            is_boundary=False)
        self.assertRaisesRegex(
            sbml.SBMLError, 'Duplicate SId "', self.r.add_modifier, species,
            sid)

    def test_products(self):

        # Check bad species
        sid = 'product'
        species = 'species'
        self.assertRaisesRegex(
            sbml.SBMLError, '<', self.r.add_product, species, sid)

        # Check invalid sid
        sid = ';'
        compartment = sbml.Compartment(self.model, sid='compartment')
        species = sbml.Species(
            compartment=compartment,
            sid='species',
            is_amount=False,
            is_constant=False,
            is_boundary=False)
        self.assertRaisesRegex(
            sbml.SBMLError, 'Invalid SId "', self.r.add_product, species, sid)

        # Check good sid
        sid = 'product'
        compartment = sbml.Compartment(self.model, sid='compartment')
        species = sbml.Species(
            compartment=compartment,
            sid='species',
            is_amount=False,
            is_constant=False,
            is_boundary=False)
        self.r.add_product(species, sid)

        self.assertEqual(len(self.r.products()), 1)
        self.assertIsInstance(
            self.r.products()[0], sbml.SpeciesReference)

        # Check duplicate sid
        sid = 'product'
        compartment = sbml.Compartment(self.model, sid='compartment')
        species = sbml.Species(
            compartment=compartment,
            sid='species',
            is_amount=False,
            is_constant=False,
            is_boundary=False)
        self.assertRaisesRegex(
            sbml.SBMLError, 'Duplicate SId "', self.r.add_product, species,
            sid)

    def test_reactants(self):

        # Check bad species
        sid = 'reactant'
        species = 'species'
        self.assertRaisesRegex(
            sbml.SBMLError, '<', self.r.add_reactant, species, sid)

        # Check invalid sid
        sid = ';'
        compartment = sbml.Compartment(self.model, sid='compartment')
        species = sbml.Species(
            compartment=compartment,
            sid='species',
            is_amount=False,
            is_constant=False,
            is_boundary=False)
        self.assertRaisesRegex(
            sbml.SBMLError, 'Invalid SId "', self.r.add_reactant, species, sid)

        # Check good sid
        sid = 'reactant'
        compartment = sbml.Compartment(self.model, sid='compartment')
        species = sbml.Species(
            compartment=compartment,
            sid='species',
            is_amount=False,
            is_constant=False,
            is_boundary=False)
        self.r.add_reactant(species, sid)

        self.assertEqual(len(self.r.reactants()), 1)
        self.assertIsInstance(
            self.r.reactants()[0], sbml.SpeciesReference)

        # Check duplicate sid
        sid = 'reactant'
        compartment = sbml.Compartment(self.model, sid='compartment')
        species = sbml.Species(
            compartment=compartment,
            sid='species',
            is_amount=False,
            is_constant=False,
            is_boundary=False)
        self.assertRaisesRegex(
            sbml.SBMLError, 'Duplicate SId "', self.r.add_reactant, species,
            sid)

    def test_sid(self):
        self.assertEqual(self.r.sid(), self.sid)

    def test_species(self):

        # Add modifier species
        sid = 'modifier_species'
        compartment = sbml.Compartment(self.model, sid='compartment')
        species = sbml.Species(
            compartment=compartment,
            sid=sid,
            is_amount=False,
            is_constant=False,
            is_boundary=False)
        self.r.add_modifier(species, sid)

        # Check that species are accessible
        self.assertIsInstance(self.r.species(sid), sbml.Species)

        # Add product species
        sid = 'product_species'
        compartment = sbml.Compartment(self.model, sid='compartment')
        species = sbml.Species(
            compartment=compartment,
            sid=sid,
            is_amount=False,
            is_constant=False,
            is_boundary=False)
        self.r.add_product(species, sid)

        # Check that species are accessible
        self.assertIsInstance(self.r.species(sid), sbml.Species)

        # Add reactant species
        sid = 'reactant_species'
        compartment = sbml.Compartment(self.model, sid='compartment')
        species = sbml.Species(
            compartment=compartment,
            sid=sid,
            is_amount=False,
            is_constant=False,
            is_boundary=False)
        self.r.add_reactant(species, sid)

        # Check that species are accessible
        self.assertIsInstance(self.r.species(sid), sbml.Species)

    def test_string_representation(self):
        self.assertEqual(str(self.r), '<Reaction ' + self.sid + '>')


class TestSpecies(unittest.TestCase):
    """
    Unit tests for :class:`Species`.
    """

    @classmethod
    def setUpClass(cls):
        cls.model = sbml.Model(name='model')
        cls.c = cls.model.add_compartment(sid='compartment')

    def test_is_amount(self):

        # Test is amount
        sid = 'species'
        species = sbml.Species(
            compartment=self.c, sid=sid, is_amount=True, is_constant=False,
            is_boundary=False)

        self.assertTrue(species.is_amount())

        # Test is concentration
        sid = 'species'
        species = sbml.Species(
            compartment=self.c, sid=sid, is_amount=False, is_constant=False,
            is_boundary=False)

        self.assertFalse(species.is_amount())

        # Test bad amount
        sid = 'species'
        self.assertRaisesRegex(
            sbml.SBMLError, 'Is_amount <', sbml.Species, self.c,
            sid, 'No', False, False)

    def test_is_boundary(self):

        # Test is boundary
        sid = 'species'
        species = sbml.Species(
            compartment=self.c, sid=sid, is_amount=False, is_constant=False,
            is_boundary=True)

        self.assertTrue(species.is_boundary())

        # Test is not boundary
        sid = 'species'
        species = sbml.Species(
            compartment=self.c, sid=sid, is_amount=False, is_constant=False,
            is_boundary=False)

        self.assertFalse(species.is_boundary())

        # Test bad boundary
        sid = 'species'
        self.assertRaisesRegex(
            sbml.SBMLError, 'Is_boundary <', sbml.Species, self.c,
            sid, False, False, 'No')

    def test_compartment(self):

        # Bad compartment
        sid = 'species'
        comp = 'compartment'

        self.assertRaisesRegex(
            sbml.SBMLError, '<', sbml.Species, comp,
            sid, False, 'No', False)

        # Test good compartment
        sid = 'species'
        comp = self.c
        species = sbml.Species(
            compartment=self.c, sid=sid, is_amount=False, is_constant=True,
            is_boundary=False)

        self.assertEqual(species.compartment(), self.c)

    def test_is_constant(self):

        # Test is constant
        sid = 'species'
        species = sbml.Species(
            compartment=self.c, sid=sid, is_amount=False, is_constant=True,
            is_boundary=False)

        self.assertTrue(species.is_constant())

        # Test is not constant
        sid = 'species'
        species = sbml.Species(
            compartment=self.c, sid=sid, is_amount=False, is_constant=False,
            is_boundary=False)

        self.assertFalse(species.is_boundary())

        # Test bad constant
        sid = 'species'
        self.assertRaisesRegex(
            sbml.SBMLError, 'Is_constant <', sbml.Species, self.c,
            sid, False, 'No', False)

    def test_conversion_factor(self):

        sid = 'species'
        species = sbml.Species(
            compartment=self.c, sid=sid, is_amount=False, is_constant=False,
            is_boundary=False)

        # Test default value
        self.assertIsNone(species.conversion_factor())

        # Bad conversion factor
        self.assertRaisesRegex(
            sbml.SBMLError, '<', species.set_conversion_factor, 10)

        # Good model conversion factor
        factor = sbml.Parameter(self.model, 'parameter')
        self.model.set_conversion_factor(factor)

        self.assertEqual(species.conversion_factor(), factor)

        # Good species conversion factor
        factor = sbml.Parameter(self.model, 'parameter 2')
        species.set_conversion_factor(factor)

        self.assertEqual(species.conversion_factor(), factor)

    def test_initial_value(self):

        sid = 'species'
        species = sbml.Species(
            compartment=self.c, sid=sid, is_amount=False, is_constant=False,
            is_boundary=False)

        # Check default initial value
        expr, is_amount = species.initial_value()
        self.assertIsNone(expr)
        self.assertIsNone(is_amount)

        # Check bad value
        expr = 2
        self.assertRaisesRegex(
            sbml.SBMLError, '<', species.set_initial_value, expr)

        # Check good value (default in_amount=False)
        expr = myokit.Number(2)
        species.set_initial_value(expr)
        value, is_amount = species.initial_value()

        self.assertEqual(value, expr)
        self.assertFalse(is_amount)

        # Check good value
        expr = myokit.Number(10.1)
        species.set_initial_value(expr, in_amount=True)
        value, is_amount = species.initial_value()

        self.assertEqual(value, expr)
        self.assertTrue(is_amount)

        # Check bad in_amount argument
        expr = myokit.Number(10.1)
        self.assertRaisesRegex(
            sbml.SBMLError,
            '<in_amount> needs to be an instance of bool or None.',
            species.set_initial_value, expr, 'Yes')

    def test_is_rate(self):

        sid = 'species'
        species = sbml.Species(
            compartment=self.c, sid=sid, is_amount=False, is_constant=False,
            is_boundary=False)

        # Check default
        self.assertFalse(species.is_rate())

        # Check setting rate to true
        expr = myokit.Number(2)
        species.set_value(value=expr, is_rate=True)

        self.assertTrue(species.is_rate())

    def test_sid(self):
        sid = 'species'
        species = sbml.Species(
            compartment=self.c, sid=sid, is_amount=False, is_constant=False,
            is_boundary=False)

        self.assertEqual(species.sid(), sid)

    def test_string_representation(self):
        sid = 'species'
        species = sbml.Species(
            compartment=self.c, sid=sid, is_amount=False, is_constant=False,
            is_boundary=False)

        self.assertEqual(str(species), '<Species ' + sid + '>')

    def test_substance_units(self):

        sid = 'species'
        species = sbml.Species(
            compartment=self.c, sid=sid, is_amount=False, is_constant=False,
            is_boundary=False)

        # Check default substance units
        unit = myokit.units.dimensionless
        self.assertEqual(species.substance_units(), unit)

        # Check bad units
        unit = 'mg'
        self.assertRaisesRegex(
            sbml.SBMLError, '<', species.set_substance_units, unit)

        # Check model substance units
        unit = myokit.units.g
        self.model.set_substance_units(unit)

        self.assertEqual(species.substance_units(), unit)

        # Check species subtance units
        unit = myokit.units.g * 1E-3
        species.set_substance_units(unit)

        self.assertEqual(species.substance_units(), unit)

    def test_value(self):

        sid = 'species'
        species = sbml.Species(
            compartment=self.c, sid=sid, is_amount=False, is_constant=False,
            is_boundary=False)

        # Check bad value
        expr = 2
        self.assertRaisesRegex(
            sbml.SBMLError, '<', species.set_value, expr)

        # Check good value
        expr = myokit.Number(2)
        species.set_value(expr)

        self.assertEqual(species.value(), expr)


class TestSpeciesReference(unittest.TestCase):
    """
    Unit tests for :class:`SpeciesReference`.
    """

    @classmethod
    def setUpClass(cls):
        model = sbml.Model(name='model')
        comp = model.add_compartment(sid='compartment')
        cls.species = sbml.Species(comp, 'species', False, False, False)
        cls.sid = 'species_reference'
        cls.sr = sbml.SpeciesReference(cls.species, cls.sid)

    def test_initial_value(self):

        # Check default initial value
        self.assertIsNone(self.sr.initial_value())

        # Check bad value
        expr = 2
        self.assertRaisesRegex(
            sbml.SBMLError, '<', self.sr.set_initial_value, expr)

        # Check good value
        expr = myokit.Number(2)
        self.sr.set_initial_value(expr)

        self.assertEqual(self.sr.initial_value(), expr)

    def test_is_rate(self):

        # Check default
        self.assertFalse(self.sr.is_rate())

        # Check setting rate to true
        expr = myokit.Number(2)
        self.sr.set_value(value=expr, is_rate=True)

        self.assertTrue(self.sr.is_rate())

    def test_sid(self):
        self.assertEqual(self.sr.sid(), self.sid)

    def test_species(self):

        # Check bad species
        species = 'species'
        self.assertRaisesRegex(
            sbml.SBMLError, '<', sbml.SpeciesReference, species)

        # Check good species
        self.assertEqual(self.sr.species(), self.species)

    def test_value(self):

        # Check bad value
        expr = 2
        self.assertRaisesRegex(
            sbml.SBMLError, '<', self.sr.set_value, expr)

        # Check good value
        expr = myokit.Number(2)
        self.sr.set_value(expr)

        self.assertEqual(self.sr.value(), expr)


class TestModifierSpeciesReference(unittest.TestCase):
    """
    Unit tests for :class:`ModifierSpeciesReference`.
    """

    @classmethod
    def setUpClass(cls):
        model = sbml.Model(name='model')
        comp = model.add_compartment(sid='compartment')
        cls.species = sbml.Species(comp, 'species', False, False, False)
        cls.sid = 'modifier_species_reference'
        cls.sr = sbml.ModifierSpeciesReference(cls.species, cls.sid)

    def test_sid(self):
        self.assertEqual(self.sr.sid(), self.sid)

    def test_species(self):

        # Check bad species
        species = 'species'
        self.assertRaisesRegex(
            sbml.SBMLError, '<', sbml.ModifierSpeciesReference, species)

        # Check good species
        self.assertEqual(self.sr.species(), self.species)


class SBMLTestMyokitModel(unittest.TestCase):
    """
    Unit tests for Model.myokit_model method.
    """

    def test_compartments(self):
        # Tests whether compartments are added properly to the myokit model.

        # Check setting initial value for size
        sm = sbml.Model()
        c = sm.add_compartment('comp')
        c.set_initial_value(myokit.Number(2))
        mm = sm.myokit_model()
        self.assertTrue(mm.has_component('comp'))
        self.assertEqual(mm.get('comp.size').rhs(), myokit.Number(2))
        self.assertFalse(mm.get('comp.size').is_state())

        # Check setting unreferenced initial value parameter for size
        sm = sbml.Model()
        p = sbml.Parameter(sm, 'parameter')
        c = sm.add_compartment('comp')
        c.set_initial_value(myokit.Name(p))
        self.assertRaisesRegex(
            sbml.SBMLError, 'Initial value for the size of <',
            sm.myokit_model)

        # Check setting referenced initial value parameter for size
        sm = sbml.Model()
        p = sm.add_parameter('parameter')
        c = sm.add_compartment('comp')
        c.set_initial_value(myokit.Name(p))
        mm = sm.myokit_model()
        self.assertTrue(mm.has_component('comp'))
        self.assertEqual(mm.get('comp.size').rhs().code(), 'global.parameter')
        self.assertFalse(mm.get('comp.size').is_state())

        # Check setting size as value
        sm = sbml.Model()
        c = sm.add_compartment('comp')
        c.set_value(myokit.Number(3))
        mm = sm.myokit_model()
        self.assertEqual(mm.get('comp.size').rhs(), myokit.Number(3))
        self.assertFalse(mm.get('comp.size').is_state())

        # Check setting unreferenced value parameter for size
        sm = sbml.Model()
        p = sbml.Parameter(sm, 'parameter')
        c = sm.add_compartment('comp')
        c.set_value(myokit.Name(p))
        self.assertRaisesRegex(
            sbml.SBMLError, 'Value for the size of <',
            sm.myokit_model)

        # Check setting referenced value parameter for size
        sm = sbml.Model()
        p = sm.add_parameter('parameter')
        c = sm.add_compartment('comp')
        c.set_value(myokit.Name(p))
        mm = sm.myokit_model()
        self.assertTrue(mm.has_component('comp'))
        self.assertEqual(mm.get('comp.size').rhs().code(), 'global.parameter')
        self.assertFalse(mm.get('comp.size').is_state())

        # Check setting size as rate
        sm = sbml.Model()
        c = sm.add_compartment('comp')
        c.set_value(myokit.Number(1.5), is_rate=True)
        with WarningCollector():
            mm = sm.myokit_model()
        self.assertEqual(mm.get('comp.size').rhs(), myokit.Number(1.5))
        self.assertTrue(mm.get('comp.size').is_state())

    def test_name(self):

        # Test regular name
        name = 'model'
        m = sbml.Model(name=name)
        m = m.myokit_model()

        self.assertEqual(m.name(), name)

        # Test name with leading underscore
        name = '_model'
        m = sbml.Model(name=name)
        m = m.myokit_model()

        self.assertEqual(m.name(), 'underscore' + name)

    def test_notes(self):

        # Test that notes are set as meta data
        n = 'These are some notes'
        m = sbml.Model()
        m.set_notes(n)
        m = m.myokit_model()

        self.assertEqual(m.meta['desc'], n)

    def test_parameters(self):
        # Tests if parameters are converted

        m = sbml.Model()
        m.add_parameter('z')
        m.add_parameter('boat')
        m.add_parameter('c')
        m = m.myokit_model()

        # Check that model created parameters in 'global' component
        self.assertTrue(m.has_variable('global.z'))
        self.assertTrue(m.has_variable('global.boat'))
        self.assertTrue(m.has_variable('global.c'))

        # Check that total number of parameters is 4 (3 parameters and time)
        self.assertEqual(m.count_variables(), 4)

    def test_parameters_initial_value(self):
        # Tests adding parameters set with initial values

        m = sbml.Model()
        p = m.add_parameter('param')
        p.set_initial_value(myokit.Number(7))
        mm = m.myokit_model()
        pp = mm.get('global.param')
        self.assertFalse(pp.is_state())
        self.assertEqual(pp.rhs(), myokit.Number(7))

        # Test units
        self.assertIsNone(pp.unit())
        p.set_units(myokit.units.pF)
        mm = m.myokit_model()
        pp = mm.get('global.param')
        self.assertFalse(pp.is_state())
        self.assertEqual(pp.rhs(), myokit.Number(7))
        self.assertEqual(pp.unit(), myokit.units.pF)

        # Unreferenced parameter
        m = sbml.Model()
        p_bad = sbml.Parameter(m, 'p_bad')
        p = m.add_parameter('param')
        p.set_initial_value(myokit.Name(p_bad))
        self.assertRaisesRegex(
            sbml.SBMLError, 'Initial value of <', m.myokit_model)

    def test_parameters_values(self):
        # Tests adding parameters set with a value (non-state)

        m = sbml.Model()
        p = m.add_parameter('param')
        p.set_value(myokit.Plus(myokit.Number(7), myokit.Number(3)))
        mm = m.myokit_model()
        pp = mm.get('global.param')
        self.assertFalse(pp.is_state())
        self.assertEqual(
            pp.rhs(), myokit.Plus(myokit.Number(7), myokit.Number(3)))

        # Test units
        self.assertIsNone(pp.unit())
        p.set_units(myokit.units.pF)
        mm = m.myokit_model()
        pp = mm.get('global.param')
        self.assertFalse(pp.is_state())
        self.assertEqual(pp.rhs().code(), '7 + 3')
        self.assertEqual(pp.unit(), myokit.units.pF)

        # Unreferenced parameter
        m = sbml.Model()
        p_bad = sbml.Parameter(m, 'p_bad')
        p = m.add_parameter('param')
        p.set_value(myokit.Name(p_bad))
        self.assertRaisesRegex(
            sbml.SBMLError, 'Value of <', m.myokit_model)

    def test_parameter_values_rate(self):
        # Tests adding parameters set with a value (state)

        m = sbml.Model()
        p = m.add_parameter('param')
        p.set_value(myokit.Number(3.2), is_rate=True)
        with WarningCollector():
            mm = m.myokit_model()
        pp = mm.get('global.param')
        self.assertTrue(pp.is_state())
        self.assertEqual(pp.rhs(), myokit.Number(3.2))
        self.assertEqual(pp.initial_value(as_float=True), 0)

        # With initial value
        p.set_initial_value(myokit.Number(1))
        mm = m.myokit_model()
        pp = mm.get('global.param')
        self.assertTrue(pp.is_state())
        self.assertEqual(pp.rhs(), myokit.Number(3.2))
        self.assertEqual(pp.initial_value(as_float=True), 1)

        # Test units
        self.assertIsNone(pp.unit())
        p.set_units(myokit.units.pF)
        mm = m.myokit_model()
        pp = mm.get('global.param')
        self.assertTrue(pp.is_state())
        self.assertEqual(pp.rhs(), myokit.Number(3.2))
        self.assertEqual(pp.unit(), myokit.units.pF)

        # Unreferenced parameter
        m = sbml.Model()
        p_bad = sbml.Parameter(m, 'p_bad')
        p = m.add_parameter('param')
        p.set_value(myokit.Name(p_bad), is_rate=True)
        with WarningCollector():
            self.assertRaisesRegex(
                sbml.SBMLError, 'Value of <', m.myokit_model)

    def test_species(self):
        # Tests whether species initialisation in amount and concentration
        # works.

        # Species in amount
        m = sbml.Model()
        c = m.add_compartment('c')
        m.add_species(c, 's1', is_amount=True)
        mm = m.myokit_model()

        # Check whether species exists in amount
        self.assertTrue(mm.has_variable('c.s1_amount'))
        self.assertFalse(mm.has_variable('c.s1_concentration'))

        # Check that component has 2 variables (compartment size and s1 amount)
        self.assertEqual(mm.get('c').count_variables(), 2)

        # Species in concentration
        m.add_species(c, 's2', is_amount=False)
        mm = m.myokit_model()

        # Check whether species exists in amount and concentration
        self.assertTrue(mm.has_variable('c.s2_amount'))
        self.assertTrue(mm.has_variable('c.s2_concentration'))

        # Check that component now has 4 variables
        self.assertEqual(mm.get('c').count_variables(), 4)

    def test_species_bad_compartment(self):
        # Tests handling of unreferenced compartment.

        m = sbml.Model()
        c = sbml.Compartment(m, 'c')
        m.add_species(c, 's1', is_amount=True)
        self.assertRaisesRegex(
            sbml.SBMLError, 'The <', m.myokit_model)

    def test_species_initial_value(self):
        # Tests converting setting species defined through an initial value

        # I: Species in amount
        m = sbml.Model()
        c = m.add_compartment('comp')
        c.set_initial_value(myokit.Number(2))
        s1 = m.add_species(c, 'spec_1', is_amount=True)

        # Initial value in amount
        s1.set_initial_value(value=myokit.Number(3), in_amount=True)
        mm = m.myokit_model()
        ms = mm.get('comp.spec_1_amount')
        self.assertFalse(ms.is_state())
        self.assertEqual(ms.rhs(), myokit.Number(3))

        # Initial value in concentration
        s1.set_initial_value(value=myokit.Number(3), in_amount=False)
        mm = m.myokit_model()
        ms = mm.get('comp.spec_1_amount')
        self.assertFalse(ms.is_state())
        self.assertEqual(
            ms.rhs().code(), '3 * comp.size')
        self.assertEqual(ms.eval(), 3 * 2)

        # Species in concentration
        s2 = m.add_species(c, 'spec_2', is_amount=False)
        s2.set_initial_value(myokit.Number(4))
        mm = m.myokit_model()
        sc = mm.get('comp.spec_2_concentration')
        self.assertFalse(sc.is_state())
        self.assertEqual(
            sc.rhs().code(), 'comp.spec_2_amount / comp.size')
        sa = mm.get('comp.spec_2_amount')
        self.assertFalse(sa.is_state())
        self.assertEqual(sa.rhs().code(), '4 * comp.size')

        # Species in amount: unreferenced parameter
        p1 = sbml.Parameter(m, 'p1')
        s1.set_initial_value(value=myokit.Name(p1), in_amount=True)
        self.assertRaisesRegex(
            sbml.SBMLError, 'Initial value of <', m.myokit_model)

        # Species in amount: referenced parameter
        p1 = m.add_parameter('p1')
        s1.set_initial_value(value=myokit.Name(p1), in_amount=True)
        mm = m.myokit_model()
        ms = mm.get('comp.spec_1_amount')
        self.assertFalse(ms.is_state())
        self.assertEqual(ms.rhs().code(), 'global.p1')

        # Species in concentration: unreferenced parameter
        p2 = sbml.Parameter(m, 'p2')
        s2.set_initial_value(value=myokit.Name(p2), in_amount=False)
        self.assertRaisesRegex(
            sbml.SBMLError, 'Initial value of <', m.myokit_model)

    def test_species_value(self):
        # Tests converting setting species defined through a normal equation

        # Species in amount
        m = sbml.Model()
        c = m.add_compartment('comp')
        s1 = m.add_species(c, 'spec_1', is_amount=True)
        s1.set_value(myokit.Number(3))
        mm = m.myokit_model()
        ms = mm.get('comp.spec_1_amount')
        self.assertFalse(ms.is_state())
        self.assertEqual(ms.rhs(), myokit.Number(3))

        # Species in concentration
        s2 = m.add_species(c, 'spec_2', is_amount=False)
        s2.set_value(myokit.Number(4))
        mm = m.myokit_model()
        sc = mm.get('comp.spec_2_concentration')
        self.assertFalse(sc.is_state())
        self.assertEqual(
            sc.rhs().code(), 'comp.spec_2_amount / comp.size')
        sa = mm.get('comp.spec_2_amount')
        self.assertFalse(sa.is_state())
        self.assertEqual(sa.rhs().code(), '4 * comp.size')

        # Species in amount: bad parameter
        m = sbml.Model()
        p = sbml.Parameter(m, 'parameter')
        c = m.add_compartment('comp')
        s1 = m.add_species(c, 'spec_1', is_amount=True)
        s1.set_value(myokit.Name(p))
        self.assertRaisesRegex(sbml.SBMLError, 'Value of <', m.myokit_model)

        # Species in concentration: bad parameter
        m = sbml.Model()
        p = sbml.Parameter(m, 'parameter')
        c = m.add_compartment('comp')
        s1 = m.add_species(c, 'spec_1', is_amount=False)
        s1.set_value(myokit.Name(p))
        self.assertRaisesRegex(sbml.SBMLError, 'Value of <', m.myokit_model)

    def test_species_value_rate(self):
        # Tests converting setting species defined through an ODE equation

        # I: No initial compartment size
        m = sbml.Model()
        c = m.add_compartment('comp')

        # Species in amount
        s1 = m.add_species(c, 'spec_1', is_amount=True)
        s1.set_value(myokit.Number(3), is_rate=True)
        with WarningCollector():
            mm = m.myokit_model()
        ms = mm.get('comp.spec_1_amount')
        self.assertTrue(ms.is_state())
        self.assertEqual(ms.rhs(), myokit.Number(3))
        self.assertEqual(ms.initial_value(as_float=True), 0)
        s1.set_initial_value(myokit.Number(7))
        mm = m.myokit_model()
        ms = mm.get('comp.spec_1_amount')
        self.assertTrue(ms.is_state())
        self.assertEqual(ms.rhs(), myokit.Number(3))
        self.assertEqual(ms.initial_value(as_float=True), 7)

        # Species in concentration
        s2 = m.add_species(c, 'spec_2', is_amount=False)
        s2.set_value(myokit.Number(4), is_rate=True)
        with WarningCollector():
            mm = m.myokit_model()
        sc = mm.get('comp.spec_2_concentration')
        self.assertFalse(sc.is_state())
        self.assertEqual(
            sc.rhs().code(), 'comp.spec_2_amount / comp.size')
        sa = mm.get('comp.spec_2_amount')
        self.assertTrue(sa.is_state())
        self.assertEqual(sa.rhs().code(), '4 * comp.size')
        self.assertEqual(sa.initial_value(as_float=True), 0)

        # I: Set compartment size
        m = sbml.Model()
        c = m.add_compartment('comp')
        c.set_initial_value(myokit.Number(2))

        # Species in amount
        s1 = m.add_species(c, 'spec_1', is_amount=True)
        s1.set_value(myokit.Number(3), is_rate=True)
        with WarningCollector():
            mm = m.myokit_model()
        ms = mm.get('comp.spec_1_amount')
        self.assertTrue(ms.is_state())
        self.assertEqual(ms.rhs(), myokit.Number(3))
        self.assertEqual(ms.initial_value(as_float=True), 0)
        s1.set_initial_value(myokit.Number(7))
        mm = m.myokit_model()
        ms = mm.get('comp.spec_1_amount')
        self.assertTrue(ms.is_state())
        self.assertEqual(ms.rhs(), myokit.Number(3))
        self.assertEqual(ms.initial_value(as_float=True), 7)

        # Species in concentration
        s2 = m.add_species(c, 'spec_2', is_amount=False)
        s2.set_value(myokit.Number(4), is_rate=True)
        with WarningCollector():
            mm = m.myokit_model()
        sc = mm.get('comp.spec_2_concentration')
        self.assertFalse(sc.is_state())
        self.assertEqual(
            sc.rhs().code(), 'comp.spec_2_amount / comp.size')
        s2.set_initial_value(myokit.Number(6))
        mm = m.myokit_model()
        sa = mm.get('comp.spec_2_amount')
        self.assertTrue(sa.is_state())
        self.assertEqual(sa.rhs().code(), '4 * comp.size')
        self.assertEqual(sa.initial_value(as_float=True), 6 * 2)

    def test_species_units(self):
        # Tests whether species units are set properly.

        # Test I: No substance nor size units provided
        m = sbml.Model()
        c = m.add_compartment('c')
        s = m.add_species(c, 'spec')
        mm = m.myokit_model()
        amount = mm.get('c.spec_amount')
        conc = mm.get('c.spec_concentration')
        self.assertEqual(amount.unit(), myokit.units.dimensionless)
        self.assertEqual(conc.unit(), myokit.units.dimensionless)

        # Test II: Substance and size units provided
        c.set_size_units(myokit.units.m)
        s.set_substance_units(myokit.units.kg)
        mm = m.myokit_model()
        amount = mm.get('c.spec_amount')
        conc = mm.get('c.spec_concentration')
        self.assertEqual(amount.unit(), myokit.units.kg)
        self.assertEqual(conc.unit(), myokit.units.kg / myokit.units.meter)

    def test_reaction_expression(self):
        # Tests whether species reaction rate expressions are set correctly.

        m = sbml.Model()
        c = m.add_compartment('c')
        c.set_initial_value(myokit.Number(1.2))
        s1 = m.add_species(c, 's1')
        s1.set_initial_value(myokit.Number(2), in_amount=True)
        s2 = m.add_species(c, 's2')
        s2.set_initial_value(myokit.Number(1.5))
        r = m.add_reaction('r')
        r.add_reactant(s1)
        r.add_product(s2)
        r.set_kinetic_law(myokit.Plus(myokit.Name(s1), myokit.Name(s2)))
        mm = m.myokit_model()

        # Check that species are state variables
        var = mm.get('c.s1_amount')
        self.assertTrue(var.is_state())

        var = mm.get('c.s2_amount')
        self.assertTrue(var.is_state())

        # Check rates
        var = mm.get('c.s1_amount')
        self.assertEqual(var.eval(), -(2 / 1.2 + 1.5))

        var = mm.get('c.s2_amount')
        self.assertEqual(var.eval(), 2 / 1.2 + 1.5)

    def test_reaction_bad_compartment(self):
        # Tests handling of unreferenced compartment for reactants/products.

        # Bad compartment for reactant
        m = sbml.Model()
        c = sbml.Compartment(m, 'compartment')
        s1 = sbml.Species(c, 's1', False, False, False)
        r = m.add_reaction('r')
        r.add_reactant(s1, 'sr1')
        self.assertRaisesRegex(sbml.SBMLError, 'The <', m.myokit_model)

        # Bad compartment for product
        m = sbml.Model()
        c = sbml.Compartment(m, 'compartment')
        s1 = sbml.Species(c, 's1', False, False, False)
        r = m.add_reaction('r')
        r.add_product(s1, 'sr1')
        self.assertRaisesRegex(sbml.SBMLError, 'The <', m.myokit_model)

    def test_reaction_bad_species(self):
        # Tests handling of unreferenced species in kinetic law.

        # Bad reactant
        m = sbml.Model()
        c = m.add_compartment('c')
        c.set_initial_value(myokit.Number(1.2))
        s1 = sbml.Species(c, 's1', False, False, False)
        r = m.add_reaction('r')
        r.add_reactant(s1)
        r.set_kinetic_law(myokit.Name(s1))
        self.assertRaisesRegex(
            sbml.SBMLError, 'Kinetic law of <', m.myokit_model)

        # Bad product
        m = sbml.Model()
        c = m.add_compartment('c')
        c.set_initial_value(myokit.Number(1.2))
        s2 = sbml.Species(c, 's2', False, False, False)
        r = m.add_reaction('r')
        r.add_product(s2)
        r.set_kinetic_law(myokit.Name(s2))
        self.assertRaisesRegex(
            sbml.SBMLError, 'Kinetic law of <', m.myokit_model)

        # Good reactant, bad modifier
        m = sbml.Model()
        c = m.add_compartment('c')
        c.set_initial_value(myokit.Number(1.2))
        s1 = m.add_species(c, 's1')
        s2 = sbml.Species(c, 's2', False, False, False)
        r = m.add_reaction('r')
        r.add_reactant(s1)
        r.add_modifier(s2)
        r.set_kinetic_law(myokit.Plus(myokit.Name(s1), myokit.Name(s2)))
        with WarningCollector():
            self.assertRaisesRegex(
                sbml.SBMLError, 'Reaction rate expression of <',
                m.myokit_model)

        # Good product, bad modifier
        m = sbml.Model()
        c = m.add_compartment('c')
        c.set_initial_value(myokit.Number(1.2))
        s1 = m.add_species(c, 's1')
        s2 = sbml.Species(c, 's2', False, False, False)
        r = m.add_reaction('r')
        r.add_product(s1)
        r.add_modifier(s2)
        r.set_kinetic_law(myokit.Plus(myokit.Name(s1), myokit.Name(s2)))
        with WarningCollector():
            self.assertRaisesRegex(
                sbml.SBMLError, 'Reaction rate expression of <',
                m.myokit_model)

    def test_reaction_bad_kinetic_law(self):
        # Tests handling of unreferenced variables in ketic law.
        m = sbml.Model()
        p = sbml.Parameter(m, 'parameter')
        c = m.add_compartment('c')
        c.set_initial_value(myokit.Number(1.2))
        s1 = m.add_species(c, 's1')
        r = m.add_reaction('r')
        r.add_reactant(s1)
        r.set_kinetic_law(myokit.Plus(myokit.Name(s1), myokit.Name(p)))

        with WarningCollector():
            self.assertRaisesRegex(
                sbml.SBMLError, 'Reaction rate expression of <',
                m.myokit_model)

    def test_reaction_no_kinetic_law(self):
        # Tests whether missing kinetic law is handled correctly.

        m = sbml.Model()
        c = m.add_compartment('c')
        c.set_initial_value(myokit.Number(1.2))
        s1 = m.add_species(c, 's1')
        s1.set_initial_value(myokit.Number(2), in_amount=True)
        s2 = m.add_species(c, 's2')
        s2.set_initial_value(myokit.Number(1.5))
        r = m.add_reaction('r')
        r.add_reactant(s1)
        r.add_product(s2)
        mm = m.myokit_model()

        # Check that species are state variables
        var = mm.get('c.s1_amount')
        self.assertFalse(var.is_state())

        var = mm.get('c.s2_amount')
        self.assertFalse(var.is_state())

        # Check rates
        var = mm.get('c.s1_amount')
        self.assertEqual(var.eval(), 2)

        var = mm.get('c.s2_amount')
        self.assertEqual(var.eval(), 1.2 * 1.5)

    def test_reaction_bad_conversion_factor(self):
        # Tests whether error is thrown when conversion factor is
        # not referenced in model.

        # Bad conversion factor for reactant
        m = sbml.Model()
        px = sbml.Parameter(m, 'x')
        px.set_value(myokit.Number(1.2))
        c = m.add_compartment('c')
        c.set_initial_value(myokit.Number(1.2))
        s1 = m.add_species(c, 's1')
        s1.set_initial_value(myokit.Number(2), in_amount=True)
        s1.set_conversion_factor(px)
        s2 = m.add_species(c, 's2')
        s2.set_initial_value(myokit.Number(1.5))
        r = m.add_reaction('r')
        r.add_reactant(s1)
        r.add_product(s2)
        r.set_kinetic_law(myokit.Plus(myokit.Name(s1), myokit.Name(s2)))

        self.assertRaisesRegex(sbml.SBMLError, 'Species <', m.myokit_model)

        # Bad conversion factor for product
        m = sbml.Model()
        px = sbml.Parameter(m, 'x')
        px.set_value(myokit.Number(1.2))
        c = m.add_compartment('c')
        c.set_initial_value(myokit.Number(1.2))
        s1 = m.add_species(c, 's1')
        s1.set_initial_value(myokit.Number(2), in_amount=True)
        s2 = m.add_species(c, 's2')
        s2.set_initial_value(myokit.Number(1.5))
        s2.set_conversion_factor(px)
        r = m.add_reaction('r')
        r.add_reactant(s1)
        r.add_product(s2)
        r.set_kinetic_law(myokit.Plus(myokit.Name(s1), myokit.Name(s2)))

        self.assertRaisesRegex(sbml.SBMLError, 'Species <', m.myokit_model)

    def test_reaction_boundary_species(self):
        # Tests whether rate of boundary species remains unaltered.

        m = sbml.Model()
        c = m.add_compartment('c')
        c.set_initial_value(myokit.Number(1.2))
        s1 = m.add_species(c, 's1')
        s1.set_initial_value(myokit.Number(2), in_amount=True)
        s2 = m.add_species(compartment=c, sid='s2', is_boundary=True)
        s2.set_initial_value(myokit.Number(2), in_amount=True)
        s3 = m.add_species(compartment=c, sid='s3', is_boundary=True)
        s3.set_initial_value(myokit.Number(1.5), in_amount=False)
        r = m.add_reaction('r')
        r.add_reactant(s1)
        r.add_reactant(s2)
        r.add_product(s3)
        r.set_kinetic_law(myokit.Plus(myokit.Name(s1), myokit.Name(s3)))
        mm = m.myokit_model()

        # Check that species are state variables
        var = mm.get('c.s1_amount')
        self.assertTrue(var.is_state())

        var = mm.get('c.s2_amount')
        self.assertFalse(var.is_state())

        var = mm.get('c.s3_amount')
        self.assertFalse(var.is_state())

        # Check rhs
        var = mm.get('c.s1_amount')
        self.assertEqual(
            var.rhs().code(), '-(c.s1_concentration + c.s3_concentration)')
        self.assertEqual(var.eval(), -(2 / 1.2 + 1.5))

        var = mm.get('c.s2_amount')
        self.assertEqual(var.eval(), 2)

        var = mm.get('c.s3_amount')
        self.assertEqual(var.eval(), 1.2 * 1.5)

    def test_reaction_conversion_factor(self):
        # Tests whether rate contributions are converted correctly.

        m = sbml.Model()
        px = m.add_parameter('x')
        px.set_value(myokit.Number(1.2))
        py = m.add_parameter('y')
        py.set_value(myokit.Number(3))
        c = m.add_compartment('c')
        c.set_initial_value(myokit.Number(1.2))
        s1 = m.add_species(c, 's1')
        s1.set_initial_value(myokit.Number(2), in_amount=True)
        s1.set_conversion_factor(px)
        s2 = m.add_species(c, 's2')
        s2.set_initial_value(myokit.Number(1.5))
        s2.set_conversion_factor(py)
        r = m.add_reaction('r')
        r.add_reactant(s1)
        r.add_product(s2)
        r.set_kinetic_law(myokit.Plus(myokit.Name(s1), myokit.Name(s2)))
        mm = m.myokit_model()

        # Check that species are state variables
        var = mm.get('c.s1_amount')
        self.assertTrue(var.is_state())

        var = mm.get('c.s2_amount')
        self.assertTrue(var.is_state())

        # Check rates
        var = mm.get('c.s1_amount')
        self.assertEqual(var.eval(), -1.2 * (2 / 1.2 + 1.5))

        var = mm.get('c.s2_amount')
        self.assertEqual(var.eval(), 3 * (2 / 1.2 + 1.5))

    def test_reaction_stoichiometry(self):
        # Tests whether stoichiometry is used in reactions correctly.

        m = sbml.Model()
        c = m.add_compartment('c')
        c.set_initial_value(myokit.Number(1.2))
        s1 = m.add_species(c, 's1')
        s1.set_initial_value(myokit.Number(2), in_amount=True)
        s2 = m.add_species(c, 's2')
        s2.set_initial_value(myokit.Number(1.5))
        r = m.add_reaction('r')
        sr1 = r.add_reactant(s1)
        sr1.set_initial_value(myokit.Number(3))
        sr2 = r.add_product(s2)
        sr2.set_initial_value(myokit.Number(2))
        r.set_kinetic_law(myokit.Plus(myokit.Name(s1), myokit.Name(s2)))
        mm = m.myokit_model()

        # Check that species are state variables
        var = mm.get('c.s1_amount')
        self.assertTrue(var.is_state())

        var = mm.get('c.s2_amount')
        self.assertTrue(var.is_state())

        # Check rates
        var = mm.get('c.s1_amount')
        self.assertEqual(var.eval(), -3 * (2 / 1.2 + 1.5))

        var = mm.get('c.s2_amount')
        self.assertEqual(var.eval(), 2 * (2 / 1.2 + 1.5))

    def test_reaction_stoichiometry_parameter(self):
        # Tests whether stoichiometry is used in reactions correctly,
        # when it's set by a parameter.

        m = sbml.Model()
        c = m.add_compartment('c')
        c.set_initial_value(myokit.Number(1.2))
        s1 = m.add_species(c, 's1')
        s1.set_initial_value(myokit.Number(2), in_amount=True)
        s2 = m.add_species(c, 's2')
        s2.set_initial_value(myokit.Number(1.5))
        r = m.add_reaction('r')
        sr1 = r.add_reactant(s1, 'sr1')
        sr1.set_initial_value(myokit.Number(3))
        sr1.set_value(myokit.Number(5))
        sr2 = r.add_product(s2, 'sr2')
        sr2.set_value(value=myokit.Number(3.82), is_rate=True)
        r.set_kinetic_law(myokit.Plus(myokit.Name(s1), myokit.Name(s2)))
        with WarningCollector():
            mm = m.myokit_model()

        # Check that species are state variables
        var = mm.get('c.s1_amount')
        self.assertTrue(var.is_state())

        var = mm.get('c.s2_amount')
        self.assertTrue(var.is_state())

        # Check whether stoichiometries are state variables
        var = mm.get('c.sr1')
        self.assertFalse(var.is_state())

        var = mm.get('c.sr2')
        self.assertTrue(var.is_state())

        # Check rates of species
        var = mm.get('c.s1_amount')
        self.assertEqual(var.eval(), -5 * (2 / 1.2 + 1.5))

        var = mm.get('c.s2_amount')
        self.assertEqual(var.eval(), 0 * (2 / 1.2 + 1.5))

        # Check values for stoichiometries
        var = mm.get('c.sr1')
        self.assertEqual(var.eval(), 5)

        var = mm.get('c.sr2')
        self.assertEqual(var.initial_value(as_float=True), 0)
        self.assertEqual(var.eval(), 3.82)

    def test_reaction_stoichiometries_exist(self):
        # Tests whether stoichiometries are created properly.

        m = sbml.Model()
        c = m.add_compartment('c')
        c.set_initial_value(myokit.Number(1.2))
        s1 = m.add_species(c, 's1')
        s1.set_initial_value(myokit.Number(2), in_amount=True)
        s2 = m.add_species(c, 's2')
        s2.set_initial_value(myokit.Number(1.5))
        r = m.add_reaction('r')
        r.add_reactant(s1, 'sr1')
        r.add_product(s2, 'sr2')
        r.set_kinetic_law(myokit.Plus(myokit.Name(s1), myokit.Name(s2)))
        mm = m.myokit_model()

        # Check that stoichiometry variables exists
        self.assertTrue(mm.has_variable('c.sr1'))
        self.assertTrue(mm.has_variable('c.sr2'))

    def test_reaction_stoichiometries_initial_value(self):
        # Tests whether initial values of stoichiometries are set properly.

        m = sbml.Model()
        c = m.add_compartment('c')
        c.set_initial_value(myokit.Number(1.2))
        s1 = m.add_species(c, 's1')
        s2 = m.add_species(c, 's2')
        r = m.add_reaction('r')
        sr1 = r.add_reactant(s1, 'sr1')
        sr1.set_initial_value(myokit.Number(2.1))
        sr2 = r.add_product(s2, 'sr2')
        sr2.set_initial_value(myokit.Number(3.5))
        r.set_kinetic_law(myokit.Plus(myokit.Name(s1), myokit.Name(s2)))
        with WarningCollector():
            mm = m.myokit_model()

        # Check that initial values are set properly
        stoich_reactant = mm.get('c.sr1')
        stoich_product = mm.get('c.sr2')

        self.assertEqual(stoich_reactant.eval(), 2.1)
        self.assertEqual(stoich_product.eval(), 3.5)

        # Bad initial value
        m = sbml.Model()
        p = sbml.Parameter(m, 'parameter')
        c = m.add_compartment('c')
        s1 = m.add_species(c, 's1')
        r = m.add_reaction('r')
        sr1 = r.add_reactant(s1, 'sr1')
        sr1.set_initial_value(myokit.Name(p))
        r.set_kinetic_law(myokit.Name(s1))
        self.assertRaisesRegex(
            sbml.SBMLError, 'Initial value of <', m.myokit_model)

    def test_reaction_stoichiometry_values(self):
        # Tests whether values of parameters are set correctly.

        m = sbml.Model()
        p = m.add_parameter('V')
        p.set_value(myokit.Number(10.23))
        c = m.add_compartment('c')
        c.set_initial_value(myokit.Number(1.2))
        s1 = m.add_species(c, 's1')
        s2 = m.add_species(c, 's2')
        r = m.add_reaction('r')
        sr1 = r.add_reactant(s1, 'sr1')
        sr1.set_initial_value(myokit.Number(2.1))
        sr1.set_value(myokit.Plus(myokit.Name(p), myokit.Number(5)))
        sr2 = r.add_product(s2, 'sr2')
        sr2.set_initial_value(myokit.Number(3.5))
        sr2.set_value(
            myokit.Minus(myokit.Name(p), myokit.Number(1)), True)
        r.set_kinetic_law(myokit.Plus(myokit.Name(s1), myokit.Name(s2)))
        with WarningCollector():
            mm = m.myokit_model()

        # Check that stoichiometries are state/ not state variables
        var = mm.get('c.sr1')
        self.assertFalse(var.is_state())

        var = mm.get('c.sr2')
        self.assertTrue(var.is_state())

        # Check value of stoichiometries
        var = mm.get('c.sr1')
        self.assertEqual(var.eval(), 15.23)

        var = mm.get('c.sr2')
        self.assertEqual(var.initial_value(as_float=True), 3.5)
        self.assertEqual(var.eval(), 9.23)

        # Bad value
        m = sbml.Model()
        p = sbml.Parameter(m, 'parameter')
        c = m.add_compartment('c')
        s1 = m.add_species(c, 's1')
        r = m.add_reaction('r')
        sr1 = r.add_reactant(s1, 'sr1')
        sr1.set_value(myokit.Name(p))
        r.set_kinetic_law(myokit.Name(s1))
        self.assertRaisesRegex(
            sbml.SBMLError, 'Value of <', m.myokit_model)

    def test_reaction_stoichiometries(self):
        # Tests whether stoichiometries are added properly to the associated
        # component.

        # Add compartment and species to myokit model
        myokit_model = myokit.Model()
        component_references = {}
        variable_references = {}
        expression_references = {}
        species_amount_references = {}

        m = sbml.Model(name='model')
        c_sid = 'compartment'
        c = m.add_compartment(c_sid)

        X.add_compartments(
            m, myokit_model, component_references, variable_references,
            expression_references)

        sid = 'species'
        s = m.add_species(c, sid)

        X.add_species(
            m, component_references, species_amount_references,
            variable_references, expression_references)

        # Case I: No stoichiometry reference
        # Add reaction to sbml.Model
        sid = 'reaction'
        r = m.add_reaction(sid)
        r_sid = None
        r.add_reactant(species=s, sid=r_sid)
        p_sid = None
        r.add_product(species=s, sid=p_sid)

        # Add stoichiometries to myokit model
        X.add_stoichiometries(
            m, component_references, variable_references,
            expression_references)

        # Check that stoichiometries do not exist in myokit model and are not
        # referenced in variable_references
        comp = component_references[c_sid]
        self.assertFalse(comp.has_variable(str(r_sid)))
        self.assertFalse(comp.has_variable(str(p_sid)))
        self.assertNotIn(r_sid, variable_references.keys())
        self.assertNotIn(p_sid, variable_references.keys())

        # Case II: Existing stoichiometry reference
        # Add reaction to sbml.Model
        sid = 'reaction_2'
        r = m.add_reaction(sid)
        r_sid = 'reactant'
        rs = r.add_reactant(species=s, sid=r_sid)
        p_sid = 'product'
        ps = r.add_product(species=s, sid=p_sid)

        # Add stoichiometries to myokit model
        X.add_stoichiometries(
            m, component_references, variable_references,
            expression_references)

        # Check that stoichiometries exist in myokit model and are
        # referenced in variable_references
        comp = component_references[c_sid]
        self.assertTrue(comp.has_variable(str(r_sid)))
        self.assertTrue(comp.has_variable(str(p_sid)))
        self.assertIn(r_sid, variable_references.keys())
        self.assertIn(p_sid, variable_references.keys())

        # Check that their units are set to dimensionless
        # and that their expressions are referenced
        unit = myokit.units.dimensionless
        var = variable_references[r_sid]
        self.assertEqual(var.unit(), unit)
        self.assertEqual(
            expression_references[myokit.Name(rs)], myokit.Name(var))
        var = variable_references[p_sid]
        self.assertEqual(var.unit(), unit)
        self.assertEqual(
            expression_references[myokit.Name(ps)], myokit.Name(var))

        # Check that stoichiometries are referenced in

    def test_time_variable_is_created(self):
        # Tests whether time variable is created.

        m = sbml.Model()
        m.set_time_units(myokit.units.ampere)
        m = m.myokit_model()

        # Check that time variable exists
        self.assertTrue(m.has_variable('global.time'))

        # Check that unit is set
        var = m.get('global.time')
        self.assertEqual(var.unit(), myokit.units.ampere)

        # Check that initial value is set
        self.assertEqual(var.rhs(), myokit.Number(0, myokit.units.ampere))

        # Chet that variable is time bound
        self.assertTrue(var.binding(), 'time')

    def test_time_variable_is_usable(self):
        # Tests that time variable can be used in equations

        # Create SBML model
        s = sbml.Model()
        s.set_time_units(myokit.units.ms)
        p = s.add_parameter('param')
        p.set_value(myokit.Plus(myokit.Number(1), myokit.Name(s.time())))
        m = s.myokit_model()
        t = m.get('global.time')
        self.assertEqual(t.unit(), myokit.units.ms)
        self.assertEqual(
            m.get('global.param').rhs(),
            myokit.Plus(myokit.Number(1), myokit.Name(m.time())))


if __name__ == '__main__':
    import warnings
    warnings.simplefilter('always')
    unittest.main()



#
# Exports to MathML based formats.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
from lxml import etree

import myokit


class HTMLExporter(myokit.formats.Exporter):
    """
    This :class:`Exporter <myokit.formats.Exporter>` generates a HTML file
    displaying a model's equations.

    The equations are encoded using Presentation MathML. This format can be
    viewed in most modern browsers, but is less suitable as an exchange format.
    """

    def model(self, path, model, protocol=None):
        """
        Export to a html document.
        """
        # Get model name
        try:
            name = model.meta['name']
        except KeyError:
            name = 'Generated model'

        # Create model html element
        html = etree.Element('html')
        head = etree.SubElement(html, 'head')
        title = etree.SubElement(head, 'title')
        title.text = name
        body = etree.SubElement(html, 'body')
        heading = etree.SubElement(body, 'h1')
        heading.text = name

        # Create expression writer
        import myokit.formats.mathml
        writer = myokit.formats.mathml.MathMLExpressionWriter()
        writer.set_mode(presentation=True)
        writer.set_time_variable(model.time())

        # Write equations, per component
        for component in model.components():
            div = etree.SubElement(body, 'div')
            div.attrib['class'] = 'component'
            heading = etree.SubElement(div, 'h2')
            heading.text = component.qname()
            for var in component.variables(deep=True):
                div2 = etree.SubElement(div, 'div')
                div2.attrib['class'] = 'variable'
                math = etree.SubElement(div2, 'math')
                math.attrib['xmlns'] = 'http://www.w3.org/1998/Math/MathML'
                writer.eq(var.eq(), math)

        # Write xml to file
        doc = etree.ElementTree(html)
        doc.write(path, encoding='utf-8', method='xml')

        # Pretty output
        if True:
            import xml.dom.minidom as m
            xml = m.parse(path)
            with open(path, 'wb') as f:
                f.write(xml.toprettyxml(encoding='utf-8'))

    def supports_model(self):
        """
        Returns ``True``.
        """
        return True



#
# Creates common graphs
# Uses matplotlib
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import numpy as np

# Don't import pyplot yet, this will cause a crash if no window environment
# is loaded.


def simulation_times(
        time=None, realtime=None, evaluations=None, mode='stair', axes=None,
        nbuckets=50, label=None):
    """
    Draws a graph of step sizes used during a simulation.

    Data can be passed in as ``time`` (simulation time) ``realtime``
    (benchmarked time during the simulation) and ``evaluations`` (the number
    of evaluations needed for each step). Which of these fields are required
    dependens on the chosen plot ``mode``:

    ``stair``
        Draws ``time`` on the x-axis, and the step number on the y-axis.
        In this plot, a high slope means the integrator is taking lots of
        steps.
    ``stair_inverse``
        Draws ``time`` on the y-axis, and the step number on the x-axis.
    ``load``
        Draws ``time`` on the x-axis, and log(1 / step size) on the y-axis.
        In this plot, high values on the y-axis should be found near difficult
        times on the x-axis
    ``histo``
        Lumps ``time`` into buckets (whose number can be specified using the
        argument ``nbuckets``) and counts the number of steps in each bucket.
        In the final result, the times corresponding to the buckets are plotted
        on the x axis and the number of evaluations in each bucket is plotted
        on the y axis.
    ``time_per_step``
        Uses the ``realtime`` argument to calculate the time taken to advance
        the solution each step.
        In the resulting plot, the step count is plotted on the x-axis, while
        the y-axis shows the time spent at this point.
    ``eval_per_step``
        Uses the ``evaluations`` entry to calculate the number of rhs
        evaluations required for each step.
        In the resulting plot, the step number is plotted on the x-axis, and
        the number of rhs evaluations for each step is plotted on the y-axis.

    The argument ``axes`` can be used to pass in a matplotlib axes object to be
    used for the plot. If none are given, the current axes obtained from
    ``pyplot.gca()`` are used.

    Returns a matplotlib axes object.
    """
    import matplotlib.pyplot as plt
    if axes is None:
        axes = plt.gca()

    def stair(ax, time, realtime, evaluations):
        if time is None:
            raise ValueError('This plotting mode requires "time" to be set.')
        time = np.array(time, copy=False)
        step = np.arange(0, len(time))
        ax.step(time, step, label=label)

    def stair_inverse(ax, time, realtime, evaluations):
        if time is None:
            raise ValueError('This plotting mode requires "time" to be set.')
        time = np.array(time, copy=False)
        step = np.arange(0, len(time))
        ax.step(step, time, label=label)

    def load(ax, time, realtime, evaluations):
        if time is None:
            raise ValueError('This plotting mode requires "time" to be set.')
        time = np.array(time, copy=False)
        size = np.log(1 / (time[1:] - time[:-1]))
        ax.step(time[1:], size, label=label)

    def histo(ax, time, realtime, evaluations):
        if time is None:
            raise ValueError('This plotting mode requires "time" to be set.')
        time = np.array(time, copy=False)
        zero = float(time[0])
        bucket_w = (time[-1] - zero) / nbuckets
        bucket_x = np.zeros(nbuckets)
        bucket_y = np.zeros(nbuckets)
        hi = zero
        for k in range(nbuckets):
            lo = hi
            hi = zero + (k + 1) * bucket_w
            bucket_x[k] = lo
            bucket_y[k] = np.sum((lo < time) * (time <= hi))
        bucket_y[0] += 1  # First bucket contains point lo == time
        ax.step(bucket_x, bucket_y, where='post', label=label)

    def time_per_step(ax, time, realtime, evaluations):
        if realtime is None:
            raise ValueError(
                'This plotting mode requires "realtime" to be set.')
        real = np.array(realtime)
        real = real[1:] - real[:-1]
        step = np.arange(1, 1 + len(real))
        ax.step(step, real, where='mid', label=label)

    def eval_per_step(ax, time, realtime, evaluations):
        if evaluations is None:
            raise ValueError(
                'This plotting mode requires "evaluations" to be set.')
        evls = np.array(evaluations)
        evls = evls[1:] - evls[:-1]
        step = np.arange(1, 1 + len(evls))
        ax.step(step, evls, where='mid', label=label)

    modes = {
        'stair': stair,
        'stair_inverse': stair_inverse,
        'load': load,
        'histo': histo,
        'time_per_step': time_per_step,
        'eval_per_step': eval_per_step,
    }

    try:
        fn = modes[mode]
    except KeyError:
        raise ValueError(
            'Selected mode not found. Avaiable modes are: '
            + ', '.join(['"' + x + '"' for x in modes.keys()]))
    return fn(axes, time, realtime, evaluations)


def current_arrows(log, voltage, currents, axes=None):
    """
    Draws a graph of voltage versus time with arrows to indicate which currents
    are active at which stage of the action potential.

    The argument, ``log`` should be a:class:`myokit.DataLog` containing
    the data needed for the plot. The argument ``voltage`` should be the key in
    ``log`` that maps to the membrane potential.

    The list ``currents`` should contain all keys of currents to display.

    Returns a matplotlib axes object.
    """
    import matplotlib.pyplot as plt

    # Get currents, normalize with respect to total current at each time
    log = log.npview()
    traces = [log[x] for x in currents]
    times = log.time()
    memv = log[voltage]

    # Get sum of _absolute_ traces!
    I_total = np.zeros(len(traces[0]))
    for I in traces:
        I_total += abs(I)

    # Create axes
    ax = axes if axes is not None else plt.gca()

    # Plot membrane potential
    ax.plot(times, memv)
    ax.set_title(voltage)

    # Get width of time steps
    steps = np.concatenate((times[0:1], times, times[-1:]))
    steps = 0.5 * steps[2:] - 0.5 * steps[0:-2]

    # Find "zero" points, points of interest
    threshold_abs = 0.1
    for ii, I in enumerate(traces):

        # Capture parts where abs(I) is greather than the threshold and the
        # sign doesn't change
        parts = []
        indices = None
        sign = (I[0] >= 0)
        for k, i in enumerate(I):
            if abs(i) < threshold_abs or sign != (i >= 0):
                # Do nothing
                if indices is not None:
                    parts.append(indices)
                    indices = None
            else:
                # Store indices
                if indices is None:
                    indices = []
                indices.append(k)
            sign = (i >= 0)
        if indices is not None:
            parts.append(indices)

        # For each part, calculate
        #  the weighted midpoint in time
        #  the total charge transferred
        #  the average current
        #  the peak current
        #  the total charge transferred / the total sum charge transferred in
        #  that same time. This last measure can be used as a secondary
        #  threshold
        for part in parts:
            q_total = 0     # Sum of charge transferred
            t_total = 0     # Total time elapsed
            s_total = 0     # Sum of all currents in this time frame
            i_peak = 0      # Max absolute current
            t_mid = 0       # Weighted midpoint in time
            for k in part:
                t_total += steps[k]
                q_total += steps[k] * I[k]
                s_total += steps[k] * I_total[k]
                t_mid += steps[k] * I[k] * times[k]
                i_peak = max(i_peak, abs(I[k]))

            # Weighted midpoint in time (weight is height * width)
            t_mid /= q_total

            # Add sign to peak current
            if sum(I) < 0:
                i_peak *= -1.0

            # Add arrow
            k = np.nonzero(times >= t_mid)[0][0]
            ars = 'rarrow'
            arx = t_mid
            if k + 1 == len(times):
                # Massive current at final point, bit weird but should handle
                # this case...
                ary = memv[k]
                arr = 0
            else:
                t1 = times[k]
                t2 = times[k + 1]
                ary = (
                    (memv[k] * (t2 - t_mid) + memv[k + 1] * (t_mid - t1))
                    / (t2 - t1))
                arr = np.arctan2(t1 - t2, memv[k + 1] - memv[k]) * 180 / np.pi
                if sum(I) > 0:
                    arr += 180
                if abs(arr) > 90:
                    arr = 180 + arr
                    ars = 'larrow'

            bbox_props = dict(
                boxstyle=ars + ',pad=0.3',
                fc='w',
                ec='black',
                lw=1)

            ax.annotate(
                currents[ii],
                xy=(arx, ary),
                ha='center',
                va='center',
                rotation=arr,
                size=14,
                bbox=bbox_props)
    return ax


def cumulative_current(
        log, currents, axes=None, labels=None, colors=None, integrate=False,
        normalize=False, max_currents=None, line_args={}, fill_args={},
        normalise=None):
    """
    Plots a number of currents, one on top of the other, with the positive and
    negative parts of the current plotted separately.

    The advantage of this type of plot is that it shows the relative size of
    each current versus the others, and gives an indication of the total
    positive and negative current in a model.

    Accepts the following arguments:

    ``log``
        A:class:`myokit.DataLog` containing all the data to plot.
    ``currents``
        A list of keys, where each key corresponds to a current stored in
        ``log``.
    ``axes``
        The matplotlib axes to create the plot on.
    ``labels``
        Can be used to pass in a list containing the label to set for each
        current.
    ``colors``
        Can be used to pass in a list containing the colors to set for each
        current.
    ``integrate``
        Set this to ``True`` to plot total carried charge instead of currents.
    ``normalize``
        Set this to ``True`` to normalize the graph at every point, so that the
        relative contribution of each current is shown.
    ``max_currents``
        Set this to any integer n to display only the first n currents, and
        group the rest together in a remainder current.
    ``line_args``
        An optional dict with keyword arguments to pass in when drawing lines.
    ``fill_args``
        An optional dict with keyword arguments to pass in when drawing shaded
        areas.

    The best results are obtained if relatively constant currents are specified
    early. Another rule of thumb is to specify the currents roughly in the
    order they appear during an AP.
    """
    import matplotlib
    import matplotlib.pyplot as plt

    # Deprecated on 2023-06-7
    if normalize is not None:
        import warnings
        warnings.warn(
            'The keyword argument `normalise` is deprecated. Please use'
            ' `normalize` instead.')
        normalize = normalise
    del normalise

    # Get axes
    if axes is None:
        axes = plt.gca()

    # Get numpy version of log
    log = log.npview()

    # Get time
    t = log.time()

    # Get currents or charges
    if integrate:
        pos = np.array([log.integrate(c) for c in currents])
    else:
        pos = np.array([log[c] for c in currents])

    # Split positive and negative
    neg = np.minimum(pos, 0)
    pos = np.maximum(pos, 0)

    # Normalize
    if normalize:
        pos /= np.maximum(np.sum(pos, axis=0), 1e-99)
        neg /= -np.minimum(np.sum(neg, axis=0), -1e-99)

    # Number of currents to show
    nc = len(currents)
    if max_currents is None or max_currents + 1 >= nc:
        show_remainder = False
    else:
        show_remainder = True
        max_currents = int(max_currents)
        nc = 1 + max_currents

    # Colors
    if colors:
        while len(colors) < nc:
            colors.extend(colors)
    else:
        # Colormap
        try:
            cmap = matplotlib.colormaps['tab20']
        except AttributeError:  # pragma: no cover
            cmap = matplotlib.cm.get_cmap(name='tab20')
        colors = [cmap(i) for i in range(nc)]

    # Line drawing keyword arguments
    if 'color' not in line_args:
        line_args['color'] = 'k'
    if 'lw' not in line_args:
        line_args['lw'] = 1

    # Plot
    op = on = 0
    for k in range(nc):
        if k == max_currents:
            # Color and label for remainder
            color = '#cccccc'
            label = 'Remainder'

            # Positive and negative remainder parts
            p = op + np.sum(pos[k:], axis=0)
            n = on + np.sum(neg[k:], axis=0)

        else:
            # Get color
            color = colors[k]

            # Get label
            if labels:
                label = labels[k]
            else:
                c = currents[k]
                if integrate:
                    label = 'Q(' + c[c.find('.') + 1:] + ')'
                else:
                    label = c[c.find('.') + 1:]

            # Get positive and negative parts
            p, n = op + pos[k], on + neg[k]

        # Plot!
        axes.fill_between(t, p, op, facecolor=color, label=label, **fill_args)
        axes.fill_between(t, n, on, facecolor=color, **fill_args)
        axes.plot(t, p, **line_args)
        axes.plot(t, n, **line_args)
        on = n
        op = p
